#
#  Copyright (c) 2015 - 2025 Intel Corporation
#  SPDX-License-Identifier: BSD-3-Clause
#
"""
GEOPM IO - Helper module for parsing/processing report and trace files.
"""


from builtins import str
from collections import OrderedDict
import os
import json
import re
import pandas
import numpy
import glob
import sys
import subprocess # nosec
import psutil
import copy
import yaml
import io
import hashlib
if os.getenv('GEOPM_USE_UNSAFE_HDF5') is not None:
    from pandas import read_hdf
else:
    def read_hdf(path, name):
        if not os.path.exists(path):
            raise IOError(f'Trace HDF5 file {path} not detected')
        raise ImportWarning('Refusing to read HDF5 format files: file format could result in arbitrary code execution. To enable "export GEOPM_USE_UNSAFE_HDF5=1"')
from distutils.spawn import find_executable
from natsort import natsorted
from . import __version__

try:
    _, os.environ['COLUMNS'] = subprocess.check_output(['stty', 'size']).decode().split()
except subprocess.CalledProcessError:
    os.environ['COLUMNS'] = "200"

pandas.set_option('display.width', int(os.environ['COLUMNS']))
pandas.set_option('display.max_colwidth', 80)
pandas.set_option('display.max_columns', 100)


class AppOutput(object):
    """The container class for all trace related data.

    This class holds the relevant objects for parsing and indexing all
    data that is output from GEOPM.  This object can be created with a
    a trace glob string that will be used
    to search ``dir_name`` for the relevant files.  If files are found
    their data will be parsed into objects for easy data access.
    Additionally a :py:class:`pandas.DataFrame` is constructed containing all of the
    trace data.  These ``DataFrame``\ s are indexed based on the version of
    GEOPM found in the files, the profile name, agent name, and the number
    of times that particular configuration has been seen by the parser
    (i.e. experiment iteration).

    Attributes:
        trace_glob: The string pattern to use to search for trace files.
        dir_name: The directory path to use when searching for files.
        verbose: A ``bool`` to control whether verbose output is printed to stdout.

    """

    def __init__(self, traces=None, dir_name='.', verbose=False, do_cache=True):
        self._traces = {}
        self._traces_df = pandas.DataFrame()
        self._all_paths = []
        self._index_tracker = IndexTracker()
        self._node_names = None
        self._region_names = None

        if traces:
            if type(traces) is list:
                trace_paths = [os.path.join(dir_name, path) for path in traces]
            else:
                trace_glob = os.path.join(dir_name, traces)
                try:
                    trace_paths = glob.glob(trace_glob)
                except TypeError:
                    raise TypeError('<geopm> geopmpy.io: AppOutput: traces must be a list of paths or a glob pattern')
                trace_paths = natsorted(trace_paths)
                if len(trace_paths) == 0:
                    raise RuntimeError('<geopm> geopmpy.io: No trace files found with pattern {}.'.format(trace_glob))

            self._all_paths.extend(trace_paths)
            self._index_tracker.reset()

            if do_cache:
                # unique cache name based on trace files in this list
                paths_str = str(trace_paths)
                try:
                    h5_id = hashlib.shake_256(paths_str.encode()).hexdigest(14)
                except AttributeError:
                    h5_id = hash(paths_str)
                trace_h5_name = 'trace_{}.h5'.format(h5_id)
                self._all_paths.append(trace_h5_name)

                # check if cache is older than traces
                if os.path.exists(trace_h5_name):
                    cache_mod_time = os.path.getmtime(trace_h5_name)
                    regen_cache = False
                    for trace_file in trace_paths:
                        mod_time = os.path.getmtime(trace_file)
                        if mod_time > cache_mod_time:
                            regen_cache = True
                    if regen_cache:
                        os.remove(trace_h5_name)

                do_load_raw = True
                try:
                    self._traces_df = read_hdf(trace_h5_name, 'trace')
                    do_load_raw = False
                    if verbose:
                        sys.stdout.write(f'Loaded traces from {trace_h5_name}.\n')
                except ImportWarning as err:
                    sys.stderr.write(f'Warning: <geopm> geopmpy.io: {err}.\n')
                except IOError as err:
                    sys.stderr.write(f'Warning: <geopm> geopmpy.io: Trace HDF5 file not detected or older than traces {err}.\n')
                if do_load_raw:
                    self.parse_traces(trace_paths, verbose)
                    # Cache traces dataframe
                    try:
                        if verbose:
                            sys.stdout.write('Generating HDF5 files... ')
                        self._traces_df.to_hdf(trace_h5_name, 'trace')
                    except ImportError as error:
                        sys.stderr.write(f'Warning: <geopm> geopmpy.io: Unable to write HDF5 file: {error}\n')

                    if verbose:
                        sys.stdout.write('Done.\n')
                        sys.stdout.flush()
            else:
                self.parse_traces(trace_paths, verbose)

    def parse_traces(self, trace_paths, verbose):
        traces_df_list = []
        fileno = 1
        filesize = 0
        for tp in trace_paths:  # Get size of all trace files
            filesize += os.stat(tp).st_size
        # Abort if traces are too large
        avail_mem = psutil.virtual_memory().available
        if filesize > avail_mem // 2:
            sys.stderr.write('Warning: <geopm> geopmpy.io: Total size of traces is greater than 50% of available memory. Parsing traces will be skipped.\n')
            return

        filesize = '{}MiB'.format(filesize // 1024 // 1024)

        for tp in trace_paths:
            if verbose:
                sys.stdout.write('\rParsing trace file {} of {} ({})... '.format(fileno, len(trace_paths), filesize))
                sys.stdout.flush()
            fileno += 1
            tt = Trace(tp)
            self.add_trace_df(tt, traces_df_list)  # Handles multiple traces per node
        if verbose:
            sys.stdout.write('Done.\n')
            sys.stdout.flush()
        if verbose:
            sys.stdout.write('Creating combined traces DF... ')
            sys.stdout.flush()
        self._traces_df = pandas.concat(traces_df_list)
        self._traces_df = self._traces_df.sort_index(ascending=True)
        if verbose:
            sys.stdout.write('Done.\n')
            sys.stdout.flush()

    def remove_files(self):
        """Deletes all files currently tracked by this object."""
        for ff in self._all_paths:
            try:
                os.remove(ff)
            except OSError:
                pass

    def add_trace_df(self, tt, traces_df_list):
        """Adds a trace ``DataFrame`` to the tracking list.

        The report tracking list is used to create the combined
        ``DataFrame`` once all reports are parsed.

        Args:
            tt: The ``Trace`` object used to extract the Trace ``DataFrame``.
                This ``DataFrame`` will be indexed and added to the
                tracking list.

        """
        tdf = tt.get_df()  # TODO: this needs numeric cols optimization
        tdf = tdf.set_index(self._index_tracker.get_multiindex(tt))
        traces_df_list.append(tdf)

    def get_trace_data(self, node_name=None):
        idx = pandas.IndexSlice
        df = self._traces_df
        if node_name is not None:
            df = df.loc[idx[:, :, :, :, node_name, :, :], ]
        return df

    def get_trace_df(self):
        """Getter for the combined ``DataFrame`` of all trace files parsed.

        This ``DataFrame`` contains all data parsed, and has a complex
        ``MultiIndex`` for accessing the unique data from each individual
        trace.  For more information on this index, see the
        ``IndexTracker`` docstring.

        Returns:
            pandas.DataFrame: Contains all parsed data.

        """
        return self._traces_df


class IndexTracker(object):
    """Tracks and uniquely identifies experiment configurations for :py:class:`pandas.DataFrame` indexing.

    This object's purpose is to examine parsed data for reports or
    traces and determine if a particular experiment configuration has
    already been tracked.  A user may run the same configuration
    repeatedly in order to prove that results are repeatable and are
    not outliers.  Since the same configuration is used many times, it
    must be tracked and counted to ensure that the unique data for
    each run can be extracted later.

    The parsed data is used to extract the following fields to build
    the tracking index tuple:

    ``(<GEOPM_VERSION>, <PROFILE_NAME>, <AGENT_NAME>, <NODE_NAME>)``

    If the tuple is not contained in the ``_run_outputs`` dict, it is
    inserted with a value of 1.  The value is incremented if the tuple
    is currently in the ``_run_outputs`` dict.  This value is used to
    uniquely identify a particular set of parsed data when the
    ``MultiIndex`` is created.

    """
    def __init__(self):
        self._run_outputs = {}

    def _check_increment(self, run_output):
        """Extracts the index tuple from the parsed data and tracks it.

        Checks to see if the current ``run_output`` has been seen before.
        If so, the count is incremented.  Otherwise it is stored as 1.

        Args:
            run_output: The ``Trace`` object to be tracked.
        """
        index = (run_output.get_version(), run_output.get_start_time(),
                 os.path.basename(run_output.get_profile_name()),
                 run_output.get_agent(), run_output.get_node_name())
        if index not in self._run_outputs:
            self._run_outputs[index] = 1
        else:
            self._run_outputs[index] += 1

    def _get_base_index(self, run_output):
        """Constructs the actual index tuple to be used to construct a
           uniquely-identifying ``MultiIndex`` for this data.

        Takes a ``run_output`` as input, and returns the unique tuple to
        identify this ``run_output`` in the ``DataFrame``.  Note that this
        method appends the current experiment iteration to the end of
        the returned tuple.  E.g.:

        >>> self._index_tracker.get_base_index(rr)
        ('0.1.1+dev365gfcda929', 'geopm_test_integration', 170,
        'static_policy', 'power_balancing', 'mr-fusion2', 1)

        Args:
            run_output: The ``Trace`` object to produce an index tuple for.

        Returns:
            Tuple: This will contain all of the index fields needed to uniquely identify this data (including the
                count of how many times this experiment has been seen.

        """
        key = (run_output.get_version(), run_output.get_start_time(),
               os.path.basename(run_output.get_profile_name()),
               run_output.get_agent(), run_output.get_node_name())

        return key + (self._run_outputs[key], )

    def get_multiindex(self, run_output):
        """Returns a ``MultiIndex`` from this ``run_output``.  Used in :py:class:`pandas.DataFrame` construction.

        This will add the current ``run_output`` to the list of tracked
        data, and return a unique muiltiindex tuple to identify this
        data in a ``DataFrame``.

        For ``Trace`` objects, the integer index of the ``DataFrame`` is
        appended to the tuple.

        Args:
            run_output: The ``Trace`` object to produce an index
                        tuple for.

        Returns:
            pandas.MultiIndex: The unique index to identify this data object.

        """
        self._check_increment(run_output)

        itl = []
        index_names = ['version', 'start_time', 'name', 'agent', 'node_name', 'iteration']

        # Trace file index
        index_names.append('index')
        for ii in range(len(run_output.get_df())):  # Append the integer index to the DataFrame index
            itl.append(self._get_base_index(run_output) + (ii, ))

        mi = pandas.MultiIndex.from_tuples(itl, names=index_names)
        return mi

    def reset(self):
        """Clears the internal tracking dictionary.

        Since only one type of data (reports OR traces) can be tracked
        at once, this is necessary to reset the object's state so a
        new type of data can be tracked.

        """
        self._run_outputs = {}


class Trace(object):
    """Creates a :py:class:`pandas.DataFrame` comprised of the trace file data.

    This object will parse both the header and the CSV data in a trace
    file.  The header identifies the uniquely-identifying configuration
    for this file which is used for later indexing purposes.

    Even though ``__getattr__()`` and ``__getitem__()`` allow this object to
    effectively be treated like a ``DataFrame``, you must use ``get_df()`` if
    you're building a list of ``DataFrame``\ s to pass to ``pandas.concat()``.
    Using the raw object in a list and calling concat will cause an
    error.

    Attributes:
        trace_path: The path to the trace file to parse.
    """
    def __init__(self, trace_path, use_agent=True):
        self._path = trace_path

        old_headers = {'time': 'TIME',
                       'epoch_count': 'EPOCH_COUNT',
                       'region_hash': 'REGION_HASH',
                       'region_hint': 'REGION_HINT',
                       'region_progress': 'REGION_PROGRESS',
                       'region_count': 'REGION_COUNT',
                       'region_runtime': 'REGION_RUNTIME',
                       'energy_package': 'CPU_ENERGY',
                       'energy_dram': 'DRAM_ENERGY',
                       'power_package': 'CPU_POWER',
                       'power_dram': 'DRAM_POWER',
                       'frequency': 'FREQUENCY',
                       'cycles_thread': 'CPU_CYCLES_THREAD',
                       'cycles_reference': 'CPU_CYCLES_REFERENCE',
                       'temperature_core': 'CPU_CORE_TEMPERATURE'}

        old_balancer_headers = {'policy_power_cap': 'POLICY_POWER_CAP',
                                'policy_step_count': 'POLICY_STEP_COUNT',
                                'policy_max_epoch_runtime': 'POLICY_MAX_EPOCH_RUNTIME',
                                'policy_power_slack': 'POLICY_POWER_SLACK',
                                'epoch_runtime': 'EPOCH_RUNTIME',
                                'power_limit': 'POWER_LIMIT',
                                'enforced_power_limit': 'ENFORCED_POWER_LIMIT'}
        old_headers.update(old_balancer_headers)

        old_governor_headers = {'power_budget': 'POWER_BUDGET'}
        old_headers.update(old_governor_headers)

        # Need to determine how many lines are in the header
        # explicitly.  We cannot use '#' as a comment character since
        # it occurs in raw MSR signal names.
        skiprows = 0
        with open(trace_path) as fid:
            for ll in fid:
                if ll.startswith('#'):
                    skiprows += 1
                else:
                    break
        column_headers = pandas.read_csv(trace_path, sep='|', skiprows=skiprows, nrows=0, encoding='utf-8').columns.tolist()
        original_headers = copy.deepcopy(column_headers)

        column_headers = [old_headers.get(ii, ii) for ii in column_headers]

        if column_headers != original_headers:
            sys.stderr.write('Warning: <geopm> geopmpy.io: Old trace file format detected. Old column headers will be forced ' \
                             'to UPPERCASE.\n')

        # region_hash and region_hint must be a string for pretty printing pandas DataFrames
        # You can force them to int64 by setting up a converter function then passing the hex string through it
        # with the read_csv call, but the number will be displayed as an integer from then on.  You'd have to convert
        # it back to a hex string to compare it with the data in the reports.
        self._df = pandas.read_csv(trace_path, sep='|', skiprows=skiprows, header=0, names=column_headers, encoding='utf-8',
                                   dtype={'REGION_HASH': 'unicode', 'REGION_HINT': 'unicode'})
        self._df.columns = list(map(str.strip, self._df[:0]))  # Strip whitespace from column names
        try:
            self._df['REGION_HASH'] = self._df['REGION_HASH'].astype('unicode').map(str.strip)  # Strip whitespace from region hashes
            self._df['REGION_HINT'] = self._df['REGION_HINT'].astype('unicode').map(str.strip)  # Strip whitespace from region hints
        except KeyError: # Hash and hint are not present in profile traces
            pass

        self._version = None
        self._start_time = None
        self._profile_name = None
        self._agent = None
        self._node_name = None
        self._use_agent = use_agent
        self._parse_header(trace_path)

    def __repr__(self):
        return self._df.__repr__()

    def __str__(self):
        return self.__repr__()

    def __getattr__(self, attr):
        """Pass through attribute requests to the underlying DataFrame.

        This allows for Trace objects to be treated like DataFrames
        for analysis.  You can do things like:

        >>> tt = geopmpy.io.Trace('170-4-balanced-minife-trace-mr-fusion5')
        >>> tt.keys()
        Index([u'region_hash', u'region_hint', u'seconds', u'pkg_energy-0', u'dram_energy-0',...

        """
        return getattr(self._df, attr)

    def __getitem__(self, key):
        """Pass through item requests to the underlying DataFrame.

        This allows standard DataFrame slicing operations to take place.

        @todo, update
        >>> tt[['region_hash', 'region_hint', 'time', 'energy_package', 'energy_dram']][:5]
                     region_hash region_hint      time  energy_package-0  energy_dram-0
        0  2305843009213693952  0.662906     106012.363770   25631.015519
        1  2305843009213693952  0.667854     106012.873718   25631.045777
        2  2305843009213693952  0.672882     106013.411621   25631.075807
        3  2305843009213693952  0.677869     106013.998108   25631.105882
        4  2305843009213693952  0.682849     106014.621704   25631.136186
        """
        return self._df.__getitem__(key)

    def _parse_header(self, trace_path):
        """Parses the configuration header out of the top of the trace file.

        Args:
            trace_path: The path to the trace file to parse.
        """
        done = False
        out = []
        with open(trace_path) as fid:
            while not done:
                ll = fid.readline()
                if ll.startswith('#'):
                    out.append(ll[1:])
                else:
                    done = True
        try:
            yaml_fd = io.StringIO(u''.join(out))
            dd = yaml.safe_load(yaml_fd)
        except yaml.parser.ParserError:
            out.insert(0, '{')
            out.append('}')
            json_str = ''.join(out)
            dd = json.loads(json_str)
        try:
            self._version = dd['geopm_version']
            self._start_time = dd['start_time']
            self._profile_name = dd['profile_name']
            if self._use_agent:
                self._agent = dd['agent']
            self._node_name = dd['node_name']
        except KeyError:
            raise SyntaxError('<geopm> geopmpy.io: Trace file header could not be parsed!')

    def get_df(self):
        return self._df

    def get_version(self):
        return self._version

    def get_start_time(self):
        return self._start_time

    def get_profile_name(self):
        return self._profile_name

    def get_agent(self):
        return self._agent

    def get_node_name(self):
        return self._node_name

    @staticmethod
    def diff_df(trace_df, column_regex, epoch=True):
        """Diff the ``DataFrame``.

        Since the counters in the trace files are monotonically
        increasing, a diff must be performed to extract the useful
        data.

        Args:
            trace_df: The ``MultiIndex``\ ed :py:class:`pandas.DataFrame` created by the
                      ``AppOutput`` class.
            column_regex: A string representing the regex search
                          pattern for the column names to diff.
            epoch: A flag to set whether or not to focus solely on
                   epoch regions.

        Returns:

            pandas.DataFrame: With the diffed columns specified by ``'column_regex'``, and an ``'elapsed_time'`` column.

        TODO:
            * Should I drop everything before the first epoch if
              ``'epoch'`` is false?

        """
        # drop_duplicates() is a workaround for #662. Duplicate data
        # rows are showing up in the trace for unmarked.
        tmp_df = trace_df.drop_duplicates()

        filtered_df = tmp_df.filter(regex=column_regex).copy()
        filtered_df['elapsed_time'] = tmp_df['time']
        if epoch:
            filtered_df['epoch_count'] = tmp_df['epoch_count']
        filtered_df = filtered_df.diff()
        # The following drops all 0's and the negative sample when traversing between 2 trace files.
        # If the epoch_count column is included, this will also drop rows occurring mid-epoch.
        filtered_df = filtered_df.loc[(filtered_df > 0).all(axis=1)]

        # Reset 'index' to be 0 to the length of the unique trace files
        traces_list = []
        for (version, start_time, name, agent, node_name, iteration), df in \
            filtered_df.groupby(level=['version', 'start_time', 'name', 'agent', 'node_name', 'iteration']):
            df = df.reset_index(level='index')
            df['index'] = pandas.Series(numpy.arange(len(df)), index=df.index)
            df = df.set_index('index', append=True)
            traces_list.append(df)

        return pandas.concat(traces_list)

    @staticmethod
    def get_median_df(trace_df, column_regex, config):
        """Extract the median experiment iteration.

        This logic calculates the sum of elapsed times for all of the
        experiment iterations for all nodes in that iteration.  It
        then extracts the ``DataFrame`` for the iteration that is closest
        to the median.  For input ``DataFrame``\ s with a single iteration,
        the single iteration is returned.

        Args:
            trace_df: The ``MultiIndex``\ ed :py:class:`pandas.DataFrame` created by the
                      ``AppOutput`` class.
            column_regex: A string representing the regex search
                          pattern for the column names to diff.
            config: The ``TraceConfig`` object being used presently.

        Returns:
            pandas.DataFrame: Containing a single experiment iteration.

        """
        diffed_trace_df = Trace.diff_df(trace_df, column_regex, config.epoch_only)

        idx = pandas.IndexSlice
        et_sums = diffed_trace_df.groupby(level=['iteration'])['elapsed_time'].sum()
        median_index = (et_sums - et_sums.median()).abs().sort_values().index[0]
        median_df = diffed_trace_df.loc[idx[:, :, :, :, :, median_index], ]
        if config.verbose:
            median_df_index = []
            median_df_index.append(median_df.index.get_level_values('version').unique()[0])
            median_df_index.append(median_df.index.get_level_values('start_time').unique()[0])
            median_df_index.append(median_df.index.get_level_values('name').unique()[0])
            median_df_index.append(median_df.index.get_level_values('agent').unique()[0])
            median_df_index.append(median_df.index.get_level_values('iteration').unique()[0])
            sys.stdout.write('Median DF index = ({})...\n'.format(' '.join(str(s) for s in median_df_index)))
            sys.stdout.flush()
        return median_df


class BenchConf(object):
    """The application configuration parameters.

    Used to hold the config data for the integration test application.
    This application allows for varying combinations of regions
    (compute, IO, or network bound), complexity, desired execution
    count, and amount of imbalance between nodes during execution.

    Attributes:
        path: The output path for this configuration file.

    """
    def __init__(self, path):
        self._path = path
        self._loop_count = 1
        self._region = []
        self._big_o = []
        self._hostname = []
        self._imbalance = []

    def __repr__(self):
        template = """\
path      : {path}
regions   : {regions}
big-o     : {big_o}
loop count: {loops}
hostnames : {hosts}
imbalance : {imbalance}
"""
        return template.format(path=self._path,
                               regions=self._region,
                               big_o=self._big_o,
                               loops=self._loop_count,
                               hosts=self._hostname,
                               imbalance=self._imbalance)

    def __str__(self):
        return self.__repr__()

    def set_loop_count(self, loop_count):
        self._loop_count = loop_count

    def append_region(self, name, big_o):
        """Appends a region to the internal list.

        Args:
            name: The string representation of the region.
            big_o: The desired complexity of the region.  This
                   affects compute, IO, or network complexity
                   depending on the type of region requested.

        """
        self._region.append(name)
        self._big_o.append(big_o)

    def append_imbalance(self, hostname, imbalance):
        """Appends imbalance to the config for a particular node.

        Args:
            hostname: The name of the node.

            imbalance: The amount of imbalance to apply to the node.
                       This is specified by a float in the range
                       ``[0,1]``.  For example, specifying a value of 0.25
                       means that this node will spend 25% more time
                       executing the work than a node would by
                       default.  Nodes not specified with imbalance
                       configurations will perform normally.

        """
        self._hostname.append(hostname)
        self._imbalance.append(imbalance)

    def get_path(self):
        return self._path

    def write(self):
        """Write the current config to a file."""
        obj = {'loop-count': self._loop_count,
               'region': self._region,
               'big-o': self._big_o}

        if (self._imbalance and self._hostname):
            obj['imbalance'] = self._imbalance
            obj['hostname'] = self._hostname

        with open(self._path, 'w') as fid:
            json.dump(obj, fid)

    def get_exec_path(self):
        # Using libtool causes sporadic issues with the Intel
        # toolchain.
        result = 'geopmbench'
        path = find_executable(result)
        source_dir = os.path.dirname(
                     os.path.dirname(
                     os.path.dirname(
                     os.path.realpath(__file__))))
        source_bin = os.path.join(source_dir, '.libs', 'geopmbench')
        if not path:
            result = source_bin
        else:
            with open(path, 'rb') as fid:
                buffer = fid.read(4096)
                if b'Generated by libtool' in buffer:
                    result = source_bin
        return result

    def get_exec_args(self):
        return [self._path]


class RawReport(object):
    def __init__(self, path):
        # Fix issue with python yaml module where it is confused
        # about floating point numbers of the form "1e+10" where
        # the decimal point is missing.
        # See PR: https://github.com/yaml/pyyaml/pull/174
        # for upstream fix to pyyaml
        yaml.SafeLoader.add_implicit_resolver(
            u'tag:yaml.org,2002:float',
            re.compile(r'''^(?:[-+]?(?:[0-9][0-9_]*)\.[0-9_]*(?:[eE][-+]?[0-9]+)?
                           |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
                           |\.[0-9_]+(?:[eE][-+]?[0-9]+)?
                           |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\.[0-9_]*
                           |[-+]?\.(?:inf|Inf|INF)
                           |\.(?:nan|NaN|NAN))$''', re.X),
            list(u'-+0123456789.'))
        with open(path) as fid:
            self._raw_dict = yaml.safe_load(fid)

    def raw_report(self):
        return copy.deepcopy(self._raw_dict)

    def dump_json(self, path):
        jdata = json.dumps(self._raw_dict)
        with open(path, 'w') as fid:
            fid.write(jdata)

    def meta_data(self):
        result = dict()
        all_keys = ['GEOPM Version',
                    'Start Time',
                    'Profile',
                    'Agent',
                    'Policy']
        for kk in all_keys:
            result[kk] = self._raw_dict[kk]
        return result

    def figure_of_merit(self):
        result = None
        try:
            result = copy.deepcopy(self._raw_dict['Figure of Merit'])
        except KeyError:
            pass
        return result

    def total_runtime(self):
        result = None
        try:
            result = copy.deepcopy(self._raw_dict['Total Runtime'])
        except KeyError:
            pass
        return result

    def host_names(self):
        return list(self._raw_dict['Hosts'].keys())

    def region_names(self, host_name):
        result = []
        try:
            result = [rr['region'] for rr in self._raw_dict['Hosts'][host_name]['Regions']]
        except KeyError:
            pass
        return result

    def raw_region(self, host_name, region_name):
        result = None
        try:
            for rr in self._raw_dict['Hosts'][host_name]['Regions']:
                if rr['region'] == region_name:
                    result = copy.deepcopy(rr)
        except KeyError:
            raise RuntimeError('report has no regions for the specified host')
        if not result:
            raise RuntimeError('region name: {} not found'.format(region_name))
        return result

    def raw_unmarked(self, host_name):
        host_data = self._raw_dict["Hosts"][host_name]
        key = 'Unmarked Totals'
        return copy.deepcopy(host_data[key])

    def raw_epoch(self, host_name):
        host_data = self._raw_dict["Hosts"][host_name]
        key = 'Epoch Totals'
        return copy.deepcopy(host_data[key])

    def raw_totals(self, host_name):
        host_data = self._raw_dict["Hosts"][host_name]
        key = 'Application Totals'
        return copy.deepcopy(host_data[key])

    def agent_host_additions(self, host_name):
        # other keys that are not region, epoch, or app
        # total i.e. from Agent::report_host()
        host_data = self._raw_dict[host_name]
        result = {}
        for key, val in host_data.items():
            if key not in ['Epoch Totals', 'Application Totals'] and not key.startswith('Region '):
                result[key] = copy.deepcopy(val)
        return result

    def get_field(self, raw_data, key, units=''):
        matches = [(len(kk), kk) for kk in raw_data if key in kk and units in kk]
        if len(matches) == 0:
            raise KeyError('<geopm> geopmpy.io: Field not found: {}'.format(key))
        match = sorted(matches)[0][1]
        return copy.deepcopy(raw_data[match])


class RawReportCollection(object):
    '''
    Used to group together a collection of related :py:class:`geopmpy.io.RawReport`\ s.
    '''

    def __init__(self, report_paths, dir_name='.', dir_cache=None, verbose=True, do_cache=True):
        self._reports_df = pandas.DataFrame()
        self._app_reports_df = pandas.DataFrame()
        self._epoch_reports_df = pandas.DataFrame()
        self._unmarked_reports_df = pandas.DataFrame()
        self._meta_data = None
        self.load_reports(report_paths, dir_name, dir_cache, verbose, do_cache)

    @staticmethod
    def make_h5_name(paths, outdir):
        paths_str = str([os.path.realpath(rr) for rr in paths])
        h5_id = hashlib.sha256(paths_str.encode()).hexdigest()[:14]
        report_h5_name = os.path.join(outdir, 'cache_{}.h5'.format(h5_id))
        return report_h5_name

    @staticmethod
    def fixup_metadata(metadata, df):
        # This block works around having mixed datatypes in the Profile column by
        # forcing the column into the NumPy S type.
        for key, val in metadata.items():
            if type(val) is not dict: # Policy is a dict and should be excluded
                df[key] = df[key].astype('S')
        return df

    def load_reports(self, reports, dir_name, dir_cache, verbose, do_cache):
        '''
        TODO:
            * copied from ``AppOutput``.  refactor to shared function.
            * removed concept of tracked files to be deleted
            * added separate epoch dataframe
        '''
        if type(reports) is list:
            report_paths = [os.path.join(dir_name, path) for path in reports]
        else:
            report_glob = os.path.join(dir_name, reports)
            try:
                report_paths = glob.glob(report_glob)
            except TypeError:
                raise TypeError('<geopm> geopmpy.io: AppOutput: reports must be a list of paths or a glob pattern')
            report_paths = natsorted(report_paths)

        if len(report_paths) == 0:
            raise RuntimeError('<geopm> geopmpy.io: No report files found with pattern {}.'.format(report_glob))

        if do_cache:
            if dir_cache is None:
                dir_cache = dir_name
            self._report_h5_name = RawReportCollection.make_h5_name(report_paths, dir_cache)
            # check if cache is older than reports
            if os.path.exists(self._report_h5_name):
                cache_mod_time = os.path.getmtime(self._report_h5_name)
                regen_cache = False
                for report_file in report_paths:
                    mod_time = os.path.getmtime(report_file)
                    if mod_time > cache_mod_time:
                        regen_cache = True
                if regen_cache:
                    os.remove(self._report_h5_name)

            try:
                if verbose:
                    sys.stdout.write('Attempting to read {}...\n'.format(self._report_h5_name))
                # load dataframes from cache
                try:
                    self._reports_df = read_hdf(self._report_h5_name, 'report')
                except KeyError:
                    pass # No regions in cached report
                self._app_reports_df = read_hdf(self._report_h5_name, 'app_report')
                try:
                    self._unmarked_reports_df = read_hdf(self._report_h5_name, 'unmarked_report')
                except KeyError:
                    pass
                try:
                    self._epoch_reports_df = read_hdf(self._report_h5_name, 'epoch_report')
                except KeyError:
                    pass
                if verbose:
                    sys.stdout.write('Loaded report data from {}.\n'.format(self._report_h5_name))
            except (IOError, ImportWarning) as err:
                if type(err) is IOError:
                    sys.stderr.write('Warning: <geopm> geopmpy.io: Report HDF5 file not detected or older than reports.\n')
                else:
                    sys.stderr.write(f'Warning: <geopm> geopmpy.io: {err}.\n')

                self.parse_reports(report_paths, verbose)

                # Cache report dataframe
                cache_created = False
                while not cache_created:
                    try:
                        if verbose:
                            sys.stdout.write('Generating HDF5 files... ')
                        self._app_reports_df.to_hdf(self._report_h5_name, 'app_report', format='table')
                        if len(self._reports_df) > 0:
                            self._reports_df.to_hdf(self._report_h5_name, 'report', format='table', append=True)
                        if len(self._unmarked_reports_df) > 0:
                            self._unmarked_reports_df.to_hdf(self._report_h5_name, 'unmarked_report', format='table', append=True)
                        if len(self._epoch_reports_df) > 0:
                            self._epoch_reports_df.to_hdf(self._report_h5_name, 'epoch_report', format='table', append=True)
                        cache_created = True
                    except TypeError as error:
                        fm = RawReportCollection.fixup_metadata
                        if verbose:
                            sys.stdout.write('Applying workaround for strings in HDF5 files... ')
                        self._reports_df = fm(self._meta_data, self._reports_df)
                        self._app_reports_df = fm(self._meta_data, self._app_reports_df)
                        self._unmarked_reports_df = fm(self._meta_data, self._unmarked_reports_df)
                        self._epoch_reports_df = fm(self._meta_data, self._epoch_reports_df)
                    except ImportError as error:
                        sys.stderr.write('Warning: <geopm> geopmpy.io: Unable to write HDF5 file: {}\n'.format(str(error)))
                        break

                if verbose:
                    sys.stdout.write('Done.\n')
                    sys.stdout.flush()

        else:
            self.parse_reports(report_paths, verbose)

    def parse_reports(self, report_paths, verbose):
        # Note: overlapping key names can break this
        # Insert repeated data for non-leaf levels
        # TODO: iteration - for now, distinguishable from start time

        def _init_tables():
            self._columns_order = {}
            self._columns_set = {}
            for name in ['region', 'unmarked', 'epoch', 'app']:
                self._columns_order[name] = []
                self._columns_set[name] = set()

        def _add_column(table_name, col_name):
            '''
            Used to add columns to the data frame in the order they appear in the report.
            '''
            if table_name not in ['region', 'unmarked', 'epoch', 'app', 'all']:
                raise RuntimeError('Invalid table name')
            if table_name == 'all':
                _add_column('region', col_name)
                _add_column('unmarked', col_name)
                _add_column('epoch', col_name)
                _add_column('app', col_name)
            elif col_name not in self._columns_set[table_name]:
                self._columns_set[table_name].add(col_name)
                self._columns_order[table_name].append(col_name)

        def _try_float(val):
            '''
            Attempt to convert values we assume are floats
            '''
            rv = val
            try:
                rv = float(val)
            except ValueError:
                pass
            return rv

        _init_tables()

        region_df_list = []
        unmarked_df_list = []
        epoch_df_list = []
        app_df_list = []

        for report in report_paths:
            if verbose:
                sys.stdout.write("Loading data from {}.\n".format(report))
            rr = RawReport(report)
            self._meta_data = rr.meta_data()
            header = {}

            # report header
            for top_key, top_val in self._meta_data.items():
                # allow one level of dict nesting in header for policy
                if type(top_val) is dict:
                    for in_key, in_val in top_val.items():
                        _add_column('all', in_key)
                        header[in_key] = _try_float(in_val)
                else:
                    _add_column('all', top_key)
                    header[top_key] = str(top_val)

            _add_column('all', 'host')
            figure_of_merit = rr.figure_of_merit()
            if figure_of_merit is not None:
                _add_column('app', 'FOM')
            total_runtime = rr.total_runtime()
            if total_runtime is not None:
                _add_column('app', 'total_runtime')
            host_names = rr.host_names()
            for host in host_names:
                # data about host to be repeated over all rows
                per_host_data = {'host': host}

                # TODO: other host data may also contain dict
                # TODO: leave out for now
                #other_host_data = rr.agent_host_additions(host)

                _add_column('region', 'region')
                region_names = rr.region_names(host)
                for region in region_names:
                    row = copy.deepcopy(header)
                    row.update(per_host_data)
                    #row['region'] = row.pop('name')
                    # TODO: region hash
                    region_data = rr.raw_region(host, region)
                    for key, val in region_data.items():
                        region_data[key] = _try_float(val)
                    row.update(region_data)
                    for cc in rr.raw_region(host, region).keys():
                        _add_column('region', cc)
                    region_df_list.append(pandas.DataFrame(row, index=[0]))

                unmarked_row = copy.deepcopy(header)
                unmarked_row.update(per_host_data)
                try:
                    unmarked_data = rr.raw_unmarked(host)
                    for key, val in unmarked_data.items():
                        unmarked_data[key] = _try_float(val)
                    for cc in unmarked_data.keys():
                        _add_column('unmarked', cc)
                    unmarked_row.update(unmarked_data)
                    unmarked_df_list.append(pandas.DataFrame(unmarked_row, index=[0]))
                except KeyError:
                    pass

                epoch_row = copy.deepcopy(header)
                epoch_row.update(per_host_data)
                try:
                    epoch_data = rr.raw_epoch(host)
                    for key, val in epoch_data.items():
                        epoch_data[key] = _try_float(val)
                    for cc in epoch_data.keys():
                        _add_column('epoch', cc)
                    epoch_row.update(epoch_data)
                    epoch_df_list.append(pandas.DataFrame(epoch_row, index=[0]))
                except KeyError:
                    pass

                app_row = copy.deepcopy(header)
                if figure_of_merit is not None:
                    app_row['FOM'] = figure_of_merit
                if total_runtime is not None:
                    app_row['total_runtime'] = total_runtime
                app_row.update(per_host_data)
                app_data = rr.raw_totals(host)
                for key, val in app_data.items():
                    app_data[key] = _try_float(val)
                for cc in app_data.keys():
                    _add_column('app', cc)
                app_row.update(app_data)
                app_df_list.append(pandas.DataFrame(app_row, index=[0]))
        # reorder the columns to order of first appearance
        try:
            df = pandas.concat(region_df_list, ignore_index=True)
            df = df.reindex(columns=self._columns_order['region'])
            self._reports_df = df
        except ValueError: # If there were no regions
            pass
        try:
            unmarked_df = pandas.concat(unmarked_df_list, ignore_index=True)
            unmarked_df = unmarked_df.reindex(columns=self._columns_order['unmarked'])
            self._unmarked_reports_df = unmarked_df
        except ValueError: # If there were no unmarked regions (non-MPI app)
            pass
        try:
            epoch_df = pandas.concat(epoch_df_list, ignore_index=True)
            epoch_df = epoch_df.reindex(columns=self._columns_order['epoch'])
            self._epoch_reports_df = epoch_df
        except (KeyError, ValueError):
            pass
        app_df = pandas.concat(app_df_list, ignore_index=True)
        app_df = app_df.reindex(columns=self._columns_order['app'])
        self._app_reports_df = app_df

    def remove_cache(self):
        try:
            os.unlink(self._report_h5_name)
        except OSError:
            pass

    # TODO: rename
    def get_df(self):
        return self._reports_df

    def get_df_filtered(self, columns):
        return self.reports_df.drop(columns=columns)

    def get_epoch_df(self):
        return self._epoch_reports_df

    def get_app_df(self):
        return self._app_reports_df

    def get_unmarked_df(self):
        return self._unmarked_reports_df



#!/usr/bin/env python3
#
#  Copyright (c) 2015 - 2025 Intel Corporation
#  SPDX-License-Identifier: BSD-3-Clause
#


import unittest
from unittest import mock
from test_helper import mock_libgeopm

mock_c = mock_libgeopm.lib

import geopmpy.policy_store

class TestPolicyStore(unittest.TestCase):
    def setUp(self):
        mock_c.reset_mock()

    def test_connect(self):
        mock_c.geopm_policystore_connect.return_value = 0
        geopmpy.policy_store.connect('qwerty')
        # Expect connect(path)
        self.assertEqual(bytearray(b'qwerty\0'),
                         mock_c.geopm_policystore_connect.call_args[0][0])

        # Expect a raised exception when there is a non-zero return code
        mock_c.geopm_policystore_connect.return_value = -1
        with self.assertRaises(RuntimeError):
            geopmpy.policy_store.connect('qwerty')

    def test_disconnect(self):
        mock_c.geopm_policystore_disconnect.return_value = 0
        geopmpy.policy_store.disconnect()

        mock_c.geopm_policystore_disconnect.return_value = -1
        with self.assertRaises(RuntimeError):
            geopmpy.policy_store.disconnect()

    def test_get_best(self):
        mock_c.geopm_policystore_get_best.return_value = 0
        geopmpy.policy_store.get_best('a1', 'p1')
        # Expect get_best(profile, agent)
        self.assertEqual(bytearray(b'a1\0'), mock_c.geopm_policystore_get_best.call_args[0][0])
        self.assertEqual(bytearray(b'p1\0'), mock_c.geopm_policystore_get_best.call_args[0][1])

        mock_c.geopm_policystore_get_best.return_value = -1
        with self.assertRaises(RuntimeError):
            geopmpy.policy_store.get_best('a1', 'p1')

    def test_set_best(self):
        mock_c.geopm_policystore_set_best.return_value = 0
        geopmpy.policy_store.set_best('a1', 'p1', [1., 2.])
        # Expect set_best(profile, agent, policy_values, default_policy)
        self.assertEqual(bytearray(b'a1\0'), mock_c.geopm_policystore_set_best.call_args[0][0])
        self.assertEqual(bytearray(b'p1\0'), mock_c.geopm_policystore_set_best.call_args[0][1])
        self.assertEqual(2, mock_c.geopm_policystore_set_best.call_args[0][2])
        self.assertEqual([1, 2], list(mock_c.geopm_policystore_set_best.call_args[0][3]))

        mock_c.geopm_policystore_set_best.return_value = -1
        with self.assertRaises(RuntimeError):
            geopmpy.policy_store.set_best('a1', 'p1', [1., 2.])

    def test_set_default(self):
        mock_c.geopm_policystore_set_default.return_value = 0
        geopmpy.policy_store.set_default('a1', [1., 2.])
        # Expect set_default(agent, policy_values, default_policy)
        self.assertEqual(bytearray(b'a1\0'), mock_c.geopm_policystore_set_default.call_args[0][0])
        self.assertEqual(2, mock_c.geopm_policystore_set_default.call_args[0][1])
        self.assertEqual([1, 2], list(mock_c.geopm_policystore_set_default.call_args[0][2]))

        mock_c.geopm_policystore_set_default.return_value = -1
        with self.assertRaises(RuntimeError):
            geopmpy.policy_store.set_default('a1', [1., 2.])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#  Copyright (c) 2015 - 2025 Intel Corporation
#  SPDX-License-Identifier: BSD-3-Clause
#
import unittest
import os
import sys

if __name__ == "__main__":
    tests_dir = os.path.dirname(os.path.abspath(__file__))
    top_dir = os.path.join(tests_dir, '..')
    loader = unittest.TestLoader()
    tests = loader.discover(start_dir=tests_dir, pattern='Test*', top_level_dir=top_dir)
    runner = unittest.TextTestRunner()
    result = runner.run(tests)
    sys.exit(not result.wasSuccessful())



#!/usr/bin/env python3
#
#  Copyright (c) 2015 - 2025 Intel Corporation
#  SPDX-License-Identifier: BSD-3-Clause
#


import unittest
import os
from unittest import mock
import shlex
try:
    # Test with str StringIO where available, and with unicode StringIO elsewhere
    from StringIO import StringIO
except ImportError:
    from io import StringIO

import geopmpy.launcher

LSCPU_OUTPUT = b'''Architecture:          x86_64
CPU op-mode(s):        32-bit, 64-bit
Byte Order:            Little Endian
CPU(s):                88
On-line CPU(s) list:   0-87
Thread(s) per core:    2
Core(s) per socket:    22
Socket(s):             2
NUMA node(s):          2
Vendor ID:             GenuineIntel
CPU family:            6
Model:                 85
Model name:            Intel(R) Xeon(R) Gold 6152 CPU @ 2.10GHz
Stepping:              4
CPU MHz:               2101.000
CPU max MHz:           2101.0000
CPU min MHz:           1000.0000
BogoMIPS:              4200.00
Virtualization:        VT-x
L1d cache:             32K
L1i cache:             32K
L2 cache:              1024K
L3 cache:              30976K
NUMA node0 CPU(s):     0-21,44-65
NUMA node1 CPU(s):     22-43,66-87
Flags:                 fpu vme de pse tsc msr pae mce cx8 apic sep mtrr pge mca cmov pat pse36 clflush dts acpi mmx fxsr sse sse2 ss ht tm pbe syscall nx pdpe1gb rdtscp lm constant_tsc art arch_perfmon pebs bts rep_good nopl xtopology nonstop_tsc aperfmperf eagerfpu pni pclmulqdq dtes64 monitor ds_cpl vmx smx est tm2 ssse3 sdbg fma cx16 xtpr pdcm pcid dca sse4_1 sse4_2 x2apic movbe popcnt tsc_deadline_timer aes xsave avx f16c rdrand lahf_lm abm 3dnowprefetch epb cat_l3 cdp_l3 intel_pt ssbd mba ibrs ibpb stibp tpr_shadow vnmi flexpriority ept vpid fsgsbase tsc_adjust bmi1 hle avx2 smep bmi2 erms invpcid rtm cqm mpx rdt_a avx512f avx512dq rdseed adx smap clflushopt clwb avx512cd avx512bw avx512vl xsaveopt xsavec xgetbv1 cqm_llc cqm_occup_llc cqm_mbm_total cqm_mbm_local dtherm ida arat pln pts hwp hwp_act_window hwp_epp hwp_pkg_req pku ospke spec_ctrl intel_stibp flush_l1d
'''
LSCPU_STDERR=b'unittest lscpu stderr\n'

SRUN_HELP = b'''Usage: srun [OPTIONS...] executable [args...]

Affinity/Multi-core options: (when the task/affinity plugin is enabled)
      --cpu-bind=             Bind tasks to CPUs
                              (see "--cpu-bind=help" for options)
'''
SRUN_STDERR=b'unittest srun stderr\n'

UNITTEST_WORKLOAD_STDOUT = b'Testing, testing, 123\n'
UNITTEST_WORKLOAD_STDERR = b'Can you hear this? Do, re, mi\n'

def mock_popen_srun(*args, **kwargs):
    arg_list = shlex.split(args[0]) if kwargs.get('shell', False) else args[0]
    if arg_list[0] == 'srun':
        mock_pid = mock.Mock(returncode=0)
        if '--help' in arg_list:
            mock_pid.communicate.return_value = (SRUN_HELP, SRUN_STDERR)
        elif 'lscpu' in arg_list:
            mock_pid.communicate.return_value = (LSCPU_OUTPUT, LSCPU_STDERR)
        elif 'unittest_workload' in arg_list:
            mock_pid.communicate.return_value = (UNITTEST_WORKLOAD_STDOUT, UNITTEST_WORKLOAD_STDERR)
        else:
            arg_string = ' '.join(arg_list)
            stdout_text = 'stdout for {}\n'.format(arg_string)
            stderr_text = 'stderr for {}\n'.format(arg_string)
            mock_pid.communicate.return_value = (stdout_text.encode(), stderr_text.encode())
        return mock_pid
    else:
        raise NotImplementedError('Popen has not been mocked for {}'.format(' '.join(arg_list)))

class TestLauncher(unittest.TestCase):
    @mock.patch('subprocess.Popen', side_effect=mock_popen_srun)
    @mock.patch('geopmpy.launcher._get_cpuset', return_value=list(range(88)))
    def test_process_count(self, mock__get_cpuset, mock_popen):
        """ Test that geopm requests an additional rank for itself by default.
        """
        launcher = geopmpy.launcher.Factory().create(
                ['unittest_geopm_launcher', 'srun', '--geopm-ctl=process', '--geopm-program-filter', 'unittest_workload', 'unittest_workload'],
                num_rank = 2, num_node = 1)
        launcher.run()
        srun_args, srun_kwargs = mock_popen.call_args

        if srun_kwargs.get('shell', False):
            srun_args = shlex.split(srun_args[0])
        else:
            srun_args = srun_args[0]

        self.assertIn('-N', srun_args)
        self.assertEqual('1', srun_args[srun_args.index('-N') + 1])

        # Expect an additional rank per node for the geopm process
        self.assertIn('-n', srun_args)
        self.assertEqual('3', srun_args[srun_args.index('-n') + 1])

    @mock.patch('subprocess.Popen', side_effect=mock_popen_srun)
    @mock.patch('geopmpy.launcher._get_cpuset', return_value=list(range(88)))
    def test_non_file_output(self, mock__get_cpuset, mock_popen):
        """ Test that the launcher can redirect stdout and stderr to a non-file writable object.
        """
        out_stream = StringIO()
        error_stream = StringIO()

        launcher = geopmpy.launcher.Factory().create(
                ['unittest_geopm_launcher', 'srun', '--geopm-ctl=process', '--geopm-program-filter', 'unittest_workload', 'unittest_workload'],
                num_rank = 2, num_node = 1)
        launcher.run(stdout=out_stream, stderr=error_stream)

        self.assertIn(UNITTEST_WORKLOAD_STDOUT.decode(), out_stream.getvalue())
        self.assertIn(UNITTEST_WORKLOAD_STDERR.decode(), error_stream.getvalue())

    @mock.patch('subprocess.Popen', side_effect=mock_popen_srun)
    @mock.patch('geopmpy.launcher._get_cpuset', return_value=list(range(88)))
    def test_main(self, mock__get_cpuset, mock_popen):
        """ Test that the geopm CLI correctly feeds ntasks and nodes to the launcher.
        """
        with mock.patch('sys.argv', ['unittest_geopm_launcher', 'srun', '--geopm-ctl=process', '--geopm-program-filter', 'unittest_workload',
                                     'unittest_workload', '--ntasks', '4', '--nodes', '2']):
            geopmpy.launcher.main()

        srun_args, srun_kwargs = mock_popen.call_args

        if srun_kwargs.get('shell', False):
            srun_args = shlex.split(srun_args[0])
        else:
            srun_args = srun_args[0]

        self.assertIn('-N', srun_args)
        self.assertEqual('2', srun_args[srun_args.index('-N') + 1])

        # Expect an additional rank per node for the geopm process
        self.assertIn('-n', srun_args)
        self.assertEqual('6', srun_args[srun_args.index('-n') + 1])

    @mock.patch('subprocess.Popen', side_effect=mock_popen_srun)
    @mock.patch('geopmpy.launcher._get_cpuset', return_value=list(range(88)))
    def test_quoted_args(self, mock__get_cpuset, mock_popen):
        """ Test that the geopm CLI correctly passes through quoted args.
        """
        workload_command = "unittest_workload 'multiple words'"
        with mock.patch('sys.argv', shlex.split(
            f'unittest_geopm_launcher srun --geopm-ctl=process --ntasks 4 --nodes 2 --geopm-program-filter unittest_workload -- {workload_command}')):
            geopmpy.launcher.main()

        # [0][0] gets the first positional arg to Popen(), which is a string
        # containing the entire srun command
        srun_popen_args = mock_popen.call_args[0][0]
        self.assertEqual(srun_popen_args[-3:], ['--', 'unittest_workload', 'multiple words'])

    @mock.patch('subprocess.Popen', side_effect=mock_popen_srun)
    @mock.patch('geopmpy.launcher._get_cpuset', return_value=list(range(88)))
    def test_affinity_disable(self, mock__get_cpuset, mock_popen):
        """ Test that geopm does not emit affinity related env options when affinity
            is disabled.
        """
        launcher = geopmpy.launcher.Factory().create(
                ['unittest_geopm_launcher', 'srun', '--geopm-ctl=process', 'unittest_workload', '--geopm-program-filter', 'unittest_workload'],
                num_rank = 2, num_node = 1)
        launcher.run()
        srun_args, srun_kwargs = mock_popen.call_args

        self.assertNotIn('--cpu-bind', srun_args[0])
        self.assertNotIn('OMP_PROC_BIND', srun_kwargs['env'].keys())

    @mock.patch('subprocess.Popen', side_effect=mock_popen_srun)
    @mock.patch('geopmpy.launcher._get_cpuset', return_value=list(range(88)))
    def test_affinity_enable(self, mock__get_cpuset, mock_popen):
        """ Test that geopm does emit affinity related env options when affinity
            is enabled.
        """
        launcher = geopmpy.launcher.Factory().create(
                ['unittest_geopm_launcher', 'srun', '--geopm-ctl=process', '--geopm-affinity-enable', '--geopm-program-filter', 'unittest_workload', 'unittest_workload'],
                num_rank = 2, num_node = 1)
        launcher.run()
        srun_args, srun_kwargs = mock_popen.call_args

        self.assertIn('--cpu-bind', srun_args[0])
        self.assertIn('OMP_PROC_BIND', srun_kwargs['env'].keys())

    @mock.patch('subprocess.Popen', side_effect=mock_popen_srun)
    @mock.patch('geopmpy.launcher._get_cpuset', return_value=list(range(2, 88)))
    def test_affinity_enable_cgroup(self, mock__get_cpuset, mock_popen):
        """ Test that geopm emits affinity options respecting cgroup constraints (CPUs 0 and 1 excluded). """
        launcher = geopmpy.launcher.Factory().create(
                ['unittest_geopm_launcher', 'srun', '--geopm-ctl=process', '--geopm-affinity-enable', '--geopm-program-filter', 'unittest_workload', 'unittest_workload'],
                num_rank = 2, num_node = 1)
        launcher.run()
        srun_args, srun_kwargs = mock_popen.call_args

        self.assertIn('--cpu-bind', srun_args[0])
        self.assertIn('OMP_PROC_BIND', srun_kwargs['env'].keys())

        cpu_bind_masks = srun_args[0][6].split(':')[1].split(',')
        self.assertGreater(len(cpu_bind_masks), 0, "No CPU masks found in the '--cpu-bind' argument.")

        disallowed_cpus = 0x3  # Binary: 0011, meaning CPUs 0 and 1 are excluded
        for mask in cpu_bind_masks:
            # Convert the mask from hex to binary and check if it excludes CPUs 0 and 1
            mask_value = int(mask, 16)
            self.assertEqual(mask_value & disallowed_cpus, 0, f"CPU mask {mask} does not exclude CPUs 0 and 1 as expected.")

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
#  Copyright (c) 2015 - 2025 Intel Corporation
#  SPDX-License-Identifier: BSD-3-Clause
#

import unittest
from unittest import mock
# Aliasing "mock_libgeopm" to keep the diff small when adding this. Alternatively,
# replace usage in this file with "mock_libgeopm.lib"
from test_helper import mock_libgeopm as _mock_libgeopm
mock_libgeopm = _mock_libgeopm.lib
import geopmpy.endpoint

class TestEndpoint(unittest.TestCase):
    def setUp(self):
        mock_libgeopm.reset()
        mock_libgeopm.geopm_endpoint_create.return_value = 0
        mock_libgeopm.geopm_endpoint_destroy.return_value = 0
        mock_libgeopm.geopm_endpoint_open.return_value = 0
        mock_libgeopm.geopm_endpoint_close.return_value = 0

        self.test_agent_name = 'my_agent'
        def mock_agent(endpoint, name_max, name_cstr):
            for idx, char in enumerate(self.test_agent_name):
                name_cstr[idx] = ord(char.encode())
            name_cstr[len(self.test_agent_name)] = ord(b'\x00')
            return 0
        mock_libgeopm.geopm_endpoint_agent.side_effect = mock_agent

        self._endpoint = geopmpy.endpoint.Endpoint('test_endpoint')

    def test_endpoint_creation_destruction(self):
        self.assertEqual("Endpoint(name='test_endpoint')", repr(self._endpoint))

        initial_destroy_count = mock_libgeopm.geopm_endpoint_destroy.call_count
        del self._endpoint
        self.assertEqual(initial_destroy_count + 1, mock_libgeopm.geopm_endpoint_destroy.call_count)

        mock_libgeopm.geopm_endpoint_create.return_value = 1
        self.assertRaises(RuntimeError, geopmpy.endpoint.Endpoint, 'test_endpoint')

    def test_endpoint_entry_exit(self):
        initial_open_count = mock_libgeopm.geopm_endpoint_open.call_count
        initial_close_count = mock_libgeopm.geopm_endpoint_close.call_count
        with self._endpoint:
            self.assertEqual(initial_open_count + 1, mock_libgeopm.geopm_endpoint_open.call_count)
            self.assertEqual(initial_close_count, mock_libgeopm.geopm_endpoint_close.call_count)
        self.assertEqual(initial_close_count + 1, mock_libgeopm.geopm_endpoint_close.call_count)

    def test_endpoint_agent_name(self):
        self.assertEqual(self.test_agent_name, self._endpoint.agent())

    def test_wait_for_agent_attach(self):
        mock_libgeopm.geopm_endpoint_wait_for_agent_attach.return_value = 1
        self.assertRaises(RuntimeError, self._endpoint.wait_for_agent_attach, 123.4)

        mock_libgeopm.geopm_endpoint_wait_for_agent_attach.return_value = 0
        self._endpoint.wait_for_agent_attach(123.4)

    def test_stop_wait_loop(self):
        mock_libgeopm.geopm_endpoint_wait_for_agent_stop_wait_loop.return_value = 1
        self.assertRaises(RuntimeError, self._endpoint.stop_wait_loop)

        mock_libgeopm.geopm_endpoint_wait_for_agent_stop_wait_loop.return_value = 0
        self._endpoint.stop_wait_loop()

    def test_reset_wait_loop(self):
        mock_libgeopm.geopm_endpoint_wait_for_agent_reset_wait_loop.return_value = 1
        self.assertRaises(RuntimeError, self._endpoint.reset_wait_loop)

        mock_libgeopm.geopm_endpoint_wait_for_agent_reset_wait_loop.return_value = 0
        self._endpoint.reset_wait_loop()

    def test_endpoint_profile_name(self):
        test_profile_name = 'my profile'

        def mock_profile_name(endpoint, name_max, name_cstr):
            for idx, char in enumerate(test_profile_name):
                name_cstr[idx] = ord(char.encode())
            name_cstr[len(test_profile_name)] = ord(b'\x00')
            return 0
        mock_libgeopm.geopm_endpoint_profile_name.side_effect = mock_profile_name
        self.assertEqual(test_profile_name, self._endpoint.profile_name())

    def test_endpoint_nodes(self):
        test_node_names = ['node 1', 'node 2']

        def mock_num_node(endpoint, num_node_p):
            num_node_p.__getitem__.return_value = len(test_node_names)
            return 0
        mock_libgeopm.geopm_endpoint_num_node.side_effect = mock_num_node

        def mock_node_name(endpoint, node_idx, name_max, name_cstr):
            for idx, char in enumerate(test_node_names[node_idx]):
                name_cstr[idx] = ord(char.encode())
            name_cstr[len(test_node_names[node_idx])] = ord(b'\x00')
            return 0
        mock_libgeopm.geopm_endpoint_node_name.side_effect = mock_node_name

        self.assertEqual(test_node_names, self._endpoint.nodes())

    def test_write_policy(self):
        test_policy = {'p0': 0, 'p1': 1}
        mock_libgeopm.geopm_endpoint_write_policy.return_value = 0
        with mock.patch('geopmpy.agent.policy_names') as policy_mock:
            policy_mock.return_value = list(test_policy)
            self._endpoint.write_policy(test_policy)
        args = mock_libgeopm.geopm_endpoint_write_policy.call_args[0]
        _, num_policy, policy_array = args
        self.assertEqual(num_policy, len(test_policy))
        self.assertEqual(policy_array[0], 0)
        self.assertEqual(policy_array[1], 1)

    def test_read_sample(self):
        test_sample = {'s0': 0, 's1': 1}
        test_age = 1.1

        def mock_read_sample(endpoint, num_sample, sample_array, sample_age_p):
            sample_array[0] = test_sample['s0']
            sample_array[1] = test_sample['s1']
            sample_age_p.__getitem__.return_value = test_age
            return 0
        mock_libgeopm.geopm_endpoint_read_sample.side_effect = mock_read_sample

        with mock.patch('geopmpy.agent.sample_names') as sample_mock:
            sample_mock.return_value = list(test_sample)
            self.assertEqual((test_age, test_sample),
                             self._endpoint.read_sample())


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
import cffi
import sys

ffibuilder = cffi.FFI()

# Check whether optional/beta headers are installed
try:
    ffibuilder.verify(r"#include <geopm_policystore.h>", optional=True)
except Exception:
    have_policystore = False
    print('Note: geopmpy failed to build a wrapper for geopm_policystore.h, which '
          'is an optional header from libgeopm. The build failure can be ignored '
          'if you do not plan to use geopmpy.policy_store.',
          file=sys.stderr)
else:
    have_policystore = True

try:
    ffibuilder.verify(r"#include <geopm_endpoint.h>", optional=True)
except Exception:
    have_endpoint = False
    print('Note: geopmpy failed to build a wrapper for geopm_endpoint.h, which '
          'is an optional header from libgeopm. The build failure can be ignored '
          'if you do not plan to use geopmpy.endpoint.',
          file=sys.stderr)
else:
    have_endpoint = True

includes = ['geopm_agent.h']
if have_policystore:
    includes.append('geopm_policystore.h')
if have_endpoint:
    includes.append('geopm_endpoint.h')

ffibuilder.set_source('_libgeopm_py_cffi',
                      '\n'.join([f'#include <{header}>' for header in includes]),
                      libraries=['geopm'])

# geopm_agent.h
ffibuilder.cdef("""
int geopm_agent_supported(const char *agent_name);

int geopm_agent_num_policy(const char *agent_name,
                           int *num_policy);

int geopm_agent_policy_name(const char *agent_name,
                            int policy_idx,
                            size_t policy_name_max,
                            char *policy_name);

int geopm_agent_policy_json(const char *agent_name,
                            const double *policy_array,
                            size_t json_string_max,
                            char *json_string);

int geopm_agent_num_sample(const char *agent_name,
                           int *num_sample);

int geopm_agent_sample_name(const char *agent_name,
                            int sample_idx,
                            size_t sample_name_max,
                            char *sample_name);

int geopm_agent_num_avail(int *num_agent);

int geopm_agent_name(int agent_idx,
                     size_t agent_name_max,
                     char *agent_name);

int geopm_agent_enforce_policy(void);
""")

if have_policystore:
    ffibuilder.cdef("""
    int geopm_policystore_connect(const char *data_path);

    int geopm_policystore_disconnect();

    int geopm_policystore_get_best(const char* agent_name, const char* profile_name,
                                   size_t max_policy_vals, double* policy_vals);

    int geopm_policystore_set_best(const char* agent_name, const char* profile_name,
                                   size_t num_policy_vals, const double* policy_vals);

    int geopm_policystore_set_default(const char* agent_name,
                                      size_t num_policy_vals, const double* policy_vals);
    """)

if have_endpoint:
    ffibuilder.cdef("""
    struct geopm_endpoint_c;

    int geopm_endpoint_create(const char *endpoint_name,
                              struct geopm_endpoint_c **endpoint);

    int geopm_endpoint_destroy(struct geopm_endpoint_c *endpoint);

    int geopm_endpoint_open(struct geopm_endpoint_c *endpoint);

    int geopm_endpoint_close(struct geopm_endpoint_c *endpoint);

    int geopm_endpoint_agent(struct geopm_endpoint_c *endpoint,
                             size_t agent_name_max,
                             char *agent_name);

    int geopm_endpoint_wait_for_agent_attach(struct geopm_endpoint_c *endpoint,
                                             double timeout);

    int geopm_endpoint_stop_wait_loop(struct geopm_endpoint_c *endpoint);

    int geopm_endpoint_reset_wait_loop(struct geopm_endpoint_c *endpoint);

    int geopm_endpoint_profile_name(struct geopm_endpoint_c *endpoint,
                                    size_t profile_name_max,
                                    char *profile_name);

    int geopm_endpoint_num_node(struct geopm_endpoint_c *endpoint,
                                int *num_node);

    int geopm_endpoint_node_name(struct geopm_endpoint_c *endpoint,
                                 int node_idx,
                                 size_t node_name_max,
                                 char *node_name);

    int geopm_endpoint_write_policy(struct geopm_endpoint_c *endpoint,
                                    size_t num_policy,
                                    const double *policy_array);

    int geopm_endpoint_read_sample(struct geopm_endpoint_c *endpoint,
                                   size_t num_sample,
                                   double *sample_array,
                                   double *sample_age_sec);
    """)

if __name__ == "__main__":
    ffibuilder.compile(verbose=True)



#  Copyright (c) 2015 - 2025 Intel Corporation
#  SPDX-License-Identifier: BSD-3-Clause
#

import os
import setuptools
import os

package_name = 'geopmpy'

script_dir = os.path.dirname(os.path.realpath(__file__))
try:
    from setuptools_scm import get_version
    version = get_version(f'{script_dir}/..')
    with open(f'{script_dir}/{package_name}/VERSION', 'w') as fid:
        fid.write(version)
except (ImportError, LookupError):
    with open(f'{script_dir}/{package_name}/VERSION', 'r') as fid:
        version = fid.read().strip()
    os.environ['SETUPTOOLS_SCM_PRETEND_VERSION'] = version

setuptools.setup(
    cffi_modules=["build_libgeopm_wrapper.py:ffibuilder"]
)




#  Copyright (c) 2015 - 2025 Intel Corporation
#  SPDX-License-Identifier: BSD-3-Clause
#

from setuptools_scm import get_version
from build import ProjectBuilder
import os
from datetime import datetime

def setup():
    script_dir = os.path.dirname(os.path.realpath(__file__))
    package_name = os.path.basename(script_dir)
    version = get_version(f'{script_dir}/..')
    with open(f'{script_dir}/{package_name}/VERSION', 'w') as fid:
        fid.write(version)
    date = datetime.today().astimezone().strftime('%a, %d %b %Y %H:%M:%S %z')
    with open(f'{script_dir}/debian/changelog.in') as fid:
       changelog = fid.read()
    changelog = changelog.replace('@VERSION@', version).replace('@DATE@', date)
    with open(f'{script_dir}/debian/changelog', 'w') as fid:
        fid.write(changelog)
    with open(f'{script_dir}/python3-geopmpy.spec.in') as fid:
        specfile = fid.read()
    archive = f'{package_name}-{version}.tar.gz'
    specfile = specfile.replace('@VERSION@', version).replace('@ARCHIVE@', archive)
    with open(f'{script_dir}/python3-geopmpy.spec', 'w') as fid:
        fid.write(specfile)
    return script_dir

def build():
    script_dir = setup()
    pb = ProjectBuilder(script_dir)
    pb.build('sdist', 'dist')

if __name__ == '__main__':
    build()



#!/usr/bin/env python3
#
#  Copyright (c) 2015 - 2025 Intel Corporation
#  SPDX-License-Identifier: BSD-3-Clause
#


import unittest
import os
import tempfile
import shutil
from unittest import mock
from collections import Counter
from contextlib import contextmanager

# test_report_cache requires the use of read_hdf() from pandas
os.environ["GEOPM_USE_UNSAFE_HDF5"] = "1"
import geopmpy.io


def touch_file(file_path):
    """ Set a file's last-modified time to now.
    Create the file if it doesn't exist.
    """
    with open(file_path, 'a'):
        os.utime(file_path, None)


@contextmanager
def self_cleaning_raw_report_collection(path):
    """Create an raw report that will remove the cache after it goes out
    of scope.

    """
    raw_report = geopmpy.io.RawReportCollection(path)
    try:
        yield raw_report
    finally:
        raw_report.remove_cache()


test_report_data_old = """\
##### geopm 1.0.0+dev30g4cccfda #####
Start Time: Thu May 30 14:38:17 2019
Profile: test_fm_stream_dgemm_mix
Agent: frequency_map
Policy: {}


Host: mcfly11
Final online freq map:
    0x000000000db5f27a: 1000000000.000000
    0x000000003a6c47e3: 1000000000.000000
    0x000000003c627f60: 1400000000.000000
    0x00000000536c798f: 1000000000.000000
    0x00000000725e8066: 2100000000.000000
    0x0000000076244144: 1100000000.000000
    0x000000009851de3f: 1000000000.000000
    0x00000000addfa74f: 1600000000.000000
    0x00000000af4cafa3: 1700000000.000000
    0x00000000ce08ae24: 1300000000.000000
    0x00000000e1242325: 1700000000.000000
    0x00000000e50a9187: 1600000000.000000
    0x00000000f9d11bbd: 1500000000.000000

Region sleep (0x00000000536c798f):
    runtime (sec): 275.998
    sync-runtime (sec): 276.581
    package-energy (joules): 33573.9
    dram-energy (joules): 2835.43
    power (watts): 121.389
    frequency (%): 50.0846
    frequency (Hz): 1e+09
    network-time (sec): 0
    count: 5500
    requested-online-frequency: 1000000000.000000
Region stream-0.35-dgemm-1.80 (0x0000000076244144):
    runtime (sec): 28.5021
    sync-runtime (sec): 28.4395
    package-energy (joules): 4804.24
    dram-energy (joules): 598.129
    power (watts): 168.928
    frequency (%): 53.9602
    frequency (Hz): 1.13317e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1100000000.000000
Region stream-0.40-dgemm-1.20 (0x000000000db5f27a):
    runtime (sec): 28.2568
    sync-runtime (sec): 28.2187
    package-energy (joules): 4575.15
    dram-energy (joules): 639.775
    power (watts): 162.132
    frequency (%): 50.303
    frequency (Hz): 1.05636e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1000000000.000000
Region stream-0.45-dgemm-0.60 (0x000000009851de3f):
    runtime (sec): 27.1528
    sync-runtime (sec): 27.1159
    package-energy (joules): 4430.24
    dram-energy (joules): 667.228
    power (watts): 163.382
    frequency (%): 50.5233
    frequency (Hz): 1.06099e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1000000000.000000
Region stream-0.30-dgemm-2.40 (0x00000000ce08ae24):
    runtime (sec): 25.6523
    sync-runtime (sec): 25.5885
    package-energy (joules): 4517.88
    dram-energy (joules): 540.39
    power (watts): 176.559
    frequency (%): 61.4774
    frequency (Hz): 1.29102e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1300000000.000000
Region stream-0.50-dgemm-0.00 (0x000000003a6c47e3):
    runtime (sec): 25.3236
    sync-runtime (sec): 25.2728
    package-energy (joules): 4308.43
    dram-energy (joules): 685.24
    power (watts): 170.477
    frequency (%): 50.8315
    frequency (Hz): 1.06746e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1000000000.000000
Region stream-0.25-dgemm-3.00 (0x000000003c627f60):
    runtime (sec): 24.4374
    sync-runtime (sec): 24.3823
    package-energy (joules): 4606.69
    dram-energy (joules): 491.049
    power (watts): 188.936
    frequency (%): 65.2223
    frequency (Hz): 1.36967e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1400000000.000000
Region stream-0.20-dgemm-3.60 (0x00000000f9d11bbd):
    runtime (sec): 23.2674
    sync-runtime (sec): 23.2105
    package-energy (joules): 4591.63
    dram-energy (joules): 443.931
    power (watts): 197.826
    frequency (%): 68.9304
    frequency (Hz): 1.44754e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1500000000.000000
Region stream-0.15-dgemm-4.20 (0x00000000addfa74f):
    runtime (sec): 20.85
    sync-runtime (sec): 20.7729
    package-energy (joules): 4424.92
    dram-energy (joules): 383.237
    power (watts): 213.014
    frequency (%): 72.3136
    frequency (Hz): 1.51858e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1600000000.000000
Region stream-0.10-dgemm-4.80 (0x00000000e50a9187):
    runtime (sec): 19.8318
    sync-runtime (sec): 19.7529
    package-energy (joules): 4429.42
    dram-energy (joules): 328.406
    power (watts): 224.242
    frequency (%): 72.2398
    frequency (Hz): 1.51704e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1600000000.000000
Region stream-0.05-dgemm-5.40 (0x00000000e1242325):
    runtime (sec): 18.2642
    sync-runtime (sec): 18.1456
    package-energy (joules): 4319.39
    dram-energy (joules): 280.249
    power (watts): 238.041
    frequency (%): 75.8382
    frequency (Hz): 1.5926e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1700000000.000000
Region stream-0.00-dgemm-6.00 (0x00000000af4cafa3):
    runtime (sec): 17.8807
    sync-runtime (sec): 17.8327
    package-energy (joules): 4367.48
    dram-energy (joules): 252.722
    power (watts): 244.914
    frequency (%): 75.7953
    frequency (Hz): 1.5917e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1700000000.000000
Region with some spaces (0x00000000af000fa3):
    runtime (sec): 17.8807
    sync-runtime (sec): 17.8327
    package-energy (joules): 4367.48
    dram-energy (joules): 252.722
    power (watts): 244.914
    frequency (%): 75.7953
    frequency (Hz): 1.5917e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1700000000.000000
Region unmarked-region (0x00000000725e8066):
    runtime (sec): 6.07865
    sync-runtime (sec): 6.13216
    package-energy (joules): 1043.75
    dram-energy (joules): 72.5349
    power (watts): 170.209
    frequency (%): 97.2362
    frequency (Hz): 2.04196e+09
    network-time (sec): 12
    count: 0
    requested-online-frequency: 2100000000.000000
Epoch Totals:
    runtime (sec): 0
    sync-runtime (sec): 0
    package-energy (joules): 0
    dram-energy (joules): 0
    power (watts): 0
    frequency (%): 0
    frequency (Hz): 0
    network-time (sec): 14
    count: 0
    epoch-runtime-ignore (sec): 0
Application Totals:
    runtime (sec): 541.507
    package-energy (joules): 83998.9
    dram-energy (joules): 8218.69
    power (watts): 155.121
    network-time (sec): 15
    ignore-time (sec): 0
    geopmctl memory HWM: 144472 kB
    geopmctl network BW (B/sec): 0.0443208

Host: mcfly12
Final online freq map:
    0x000000000db5f27a: 1000000000.000000
    0x000000003a6c47e3: 1000000000.000000
    0x000000003c627f60: 1400000000.000000
    0x00000000536c798f: 1000000000.000000
    0x00000000725e8066: 2200000000.000000
    0x0000000076244144: 1100000000.000000
    0x000000009851de3f: 1000000000.000000
    0x00000000addfa74f: 1500000000.000000
    0x00000000af4cafa3: 1700000000.000000
    0x00000000ce08ae24: 1300000000.000000
    0x00000000e1242325: 1700000000.000000
    0x00000000e50a9187: 1600000000.000000
    0x00000000f9d11bbd: 1500000000.000000

Region sleep (0x00000000536c798f):
    runtime (sec): 275.837
    sync-runtime (sec): 276.56
    package-energy (joules): 34545.2
    dram-energy (joules): 3811.41
    power (watts): 124.91
    frequency (%): 50.0117
    frequency (Hz): 1.05024e+09
    network-time (sec): 0
    count: 5500
    requested-online-frequency: 1000000000.000000
Region stream-0.35-dgemm-1.80 (0x0000000076244144):
    runtime (sec): 29.1273
    sync-runtime (sec): 29.0865
    package-energy (joules): 4968.14
    dram-energy (joules): 713.884
    power (watts): 170.806
    frequency (%): 53.9922
    frequency (Hz): 1.13384e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1100000000.000000
Region stream-0.40-dgemm-1.20 (0x000000000db5f27a):
    runtime (sec): 28.4557
    sync-runtime (sec): 28.3904
    package-energy (joules): 4697.92
    dram-energy (joules): 745.835
    power (watts): 165.476
    frequency (%): 50.3009
    frequency (Hz): 1.05632e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1000000000.000000
Region stream-0.45-dgemm-0.60 (0x000000009851de3f):
    runtime (sec): 27.2367
    sync-runtime (sec): 27.1905
    package-energy (joules): 4478.41
    dram-energy (joules): 767.496
    power (watts): 164.705
    frequency (%): 50.3711
    frequency (Hz): 1.05779e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1000000000.000000
Region stream-0.30-dgemm-2.40 (0x00000000ce08ae24):
    runtime (sec): 25.8974
    sync-runtime (sec): 25.8082
    package-energy (joules): 4672.56
    dram-energy (joules): 636.411
    power (watts): 181.05
    frequency (%): 61.493
    frequency (Hz): 1.29135e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1300000000.000000
Region stream-0.50-dgemm-0.00 (0x000000003a6c47e3):
    runtime (sec): 25.2686
    sync-runtime (sec): 25.2523
    package-energy (joules): 4412.83
    dram-energy (joules): 776.741
    power (watts): 174.75
    frequency (%): 50.9256
    frequency (Hz): 1.06944e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1000000000.000000
Region stream-0.25-dgemm-3.00 (0x000000003c627f60):
    runtime (sec): 24.5979
    sync-runtime (sec): 24.5117
    package-energy (joules): 4732.35
    dram-energy (joules): 583.029
    power (watts): 193.065
    frequency (%): 65.2297
    frequency (Hz): 1.36982e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1400000000.000000
Region stream-0.20-dgemm-3.60 (0x00000000f9d11bbd):
    runtime (sec): 23.2737
    sync-runtime (sec): 23.2129
    package-energy (joules): 4694.3
    dram-energy (joules): 527.13
    power (watts): 202.228
    frequency (%): 68.9091
    frequency (Hz): 1.44709e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1500000000.000000
Region stream-0.15-dgemm-4.20 (0x00000000addfa74f):
    runtime (sec): 22.3063
    sync-runtime (sec): 22.1766
    package-energy (joules): 4650.85
    dram-energy (joules): 466.899
    power (watts): 209.719
    frequency (%): 68.7558
    frequency (Hz): 1.44387e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1500000000.000000
Region stream-0.10-dgemm-4.80 (0x00000000e50a9187):
    runtime (sec): 20.3869
    sync-runtime (sec): 20.3101
    package-energy (joules): 4646.27
    dram-energy (joules): 410.599
    power (watts): 228.767
    frequency (%): 72.2666
    frequency (Hz): 1.5176e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1600000000.000000
Region stream-0.05-dgemm-5.40 (0x00000000e1242325):
    runtime (sec): 19.2864
    sync-runtime (sec): 19.1575
    package-energy (joules): 4618.09
    dram-energy (joules): 352.889
    power (watts): 241.059
    frequency (%): 76.1363
    frequency (Hz): 1.59886e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1700000000.000000
Region stream-0.00-dgemm-6.00 (0x00000000af4cafa3):
    runtime (sec): 18.6718
    sync-runtime (sec): 18.6886
    package-energy (joules): 4600.85
    dram-energy (joules): 321.137
    power (watts): 246.185
    frequency (%): 75.8795
    frequency (Hz): 1.59347e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1700000000.000000
Region with some spaces (0x00000000af000fa3):
    runtime (sec): 17.8807
    sync-runtime (sec): 17.8327
    package-energy (joules): 4367.48
    dram-energy (joules): 252.722
    power (watts): 244.914
    frequency (%): 75.7953
    frequency (Hz): 1.5917e+09
    network-time (sec): 0
    count: 500
    requested-online-frequency: 1700000000.000000
Region unmarked-region (0x00000000725e8066):
    runtime (sec): 1.15034
    sync-runtime (sec): 1.11497
    package-energy (joules): 226.452
    dram-energy (joules): 28.9149
    power (watts): 203.102
    frequency (%): 95.0914
    frequency (Hz): 1.99692e+09
    network-time (sec): 12
    count: 0
    requested-online-frequency: 2200000000.000000
Epoch Totals:
    runtime (sec): 0
    sync-runtime (sec): 0
    package-energy (joules): 0
    dram-energy (joules): 0
    power (watts): 0
    frequency (%): 0
    frequency (Hz): 0
    network-time (sec): 14
    count: 0
    epoch-runtime-ignore (sec): 0
Application Totals:
    runtime (sec): 541.505
    package-energy (joules): 85946.8
    dram-energy (joules): 10142.6
    power (watts): 158.718
    network-time (sec): 15
    ignore-time (sec): 0
    geopmctl memory HWM: 142300 kB
    geopmctl network BW (B/sec): 0
"""

test_report_data = """\
GEOPM Version: 1.0.0+dev30g4cccfda
Start Time: Thu May 30 14:38:17 2019
Profile: test_fm_stream_dgemm_mix
Agent: frequency_map
Policy: {}

Hosts:
  mcfly11:
    Final online freq map:
      0x0db5f27a: 1000000000.000000
      0x3a6c47e3: 1000000000.000000
      0x3c627f60: 1400000000.000000
      0x536c798f: 1000000000.000000
      0x725e8066: 2100000000.000000
      0x76244144: 1100000000.000000
      0x9851de3f: 1000000000.000000
      0xaddfa74f: 1600000000.000000
      0xaf4cafa3: 1700000000.000000
      0xce08ae24: 1300000000.000000
      0xe1242325: 1700000000.000000
      0xe50a9187: 1600000000.000000
      0xf9d11bbd: 1500000000.000000

    Regions:
    -
      region: "sleep"
      hash: 0x536c798f
      runtime (s): 275.998
      sync-runtime (s): 276.581
      package-energy (J): 33573.9
      dram-energy (J): 2835.43
      power (W): 121.389
      frequency (%): 50.0846
      frequency (Hz): 1e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 5500
      requested-online-frequency: 1000000000.000000
    -
      region: "stream-0.35-dgemm-1.80"
      hash: 0x76244144
      runtime (s): 28.5021
      sync-runtime (s): 28.4395
      package-energy (J): 4804.24
      dram-energy (J): 598.129
      power (W): 168.928
      frequency (%): 53.9602
      frequency (Hz): 1.13317e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1100000000.000000
    -
      region: "stream-0.40-dgemm-1.20"
      hash: 0x0db5f27a
      runtime (s): 28.2568
      sync-runtime (s): 28.2187
      package-energy (J): 4575.15
      dram-energy (J): 639.775
      power (W): 162.132
      frequency (%): 50.303
      frequency (Hz): 1.05636e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1000000000.000000
    -
      region: "stream-0.45-dgemm-0.60"
      hash: 0x9851de3f
      runtime (s): 27.1528
      sync-runtime (s): 27.1159
      package-energy (J): 4430.24
      dram-energy (J): 667.228
      power (W): 163.382
      frequency (%): 50.5233
      frequency (Hz): 1.06099e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1000000000.000000
    -
      region: "stream-0.30-dgemm-2.40"
      hash: 0xce08ae24
      runtime (s): 25.6523
      sync-runtime (s): 25.5885
      package-energy (J): 4517.88
      dram-energy (J): 540.39
      power (W): 176.559
      frequency (%): 61.4774
      frequency (Hz): 1.29102e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1300000000.000000
    -
      region: "stream-0.50-dgemm-0.00"
      hash: 0x3a6c47e3
      runtime (s): 25.3236
      sync-runtime (s): 25.2728
      package-energy (J): 4308.43
      dram-energy (J): 685.24
      power (W): 170.477
      frequency (%): 50.8315
      frequency (Hz): 1.06746e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1000000000.000000
    -
      region: "stream-0.25-dgemm-3.00"
      hash: 0x3c627f60
      runtime (s): 24.4374
      sync-runtime (s): 24.3823
      package-energy (J): 4606.69
      dram-energy (J): 491.049
      power (W): 188.936
      frequency (%): 65.2223
      frequency (Hz): 1.36967e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1400000000.000000
    -
      region: "stream-0.20-dgemm-3.60"
      hash: 0xf9d11bbd
      runtime (s): 23.2674
      sync-runtime (s): 23.2105
      package-energy (J): 4591.63
      dram-energy (J): 443.931
      power (W): 197.826
      frequency (%): 68.9304
      frequency (Hz): 1.44754e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1500000000.000000
    -
      region: "stream-0.15-dgemm-4.20"
      hash: 0xaddfa74f
      runtime (s): 20.85
      sync-runtime (s): 20.7729
      package-energy (J): 4424.92
      dram-energy (J): 383.237
      power (W): 213.014
      frequency (%): 72.3136
      frequency (Hz): 1.51858e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1600000000.000000
    -
      region: "stream-0.10-dgemm-4.80"
      hash: 0xe50a9187
      runtime (s): 19.8318
      sync-runtime (s): 19.7529
      package-energy (J): 4429.42
      dram-energy (J): 328.406
      power (W): 224.242
      frequency (%): 72.2398
      frequency (Hz): 1.51704e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1600000000.000000
    -
      region: "stream-0.05-dgemm-5.40"
      hash: 0xe1242325
      runtime (s): 18.2642
      sync-runtime (s): 18.1456
      package-energy (J): 4319.39
      dram-energy (J): 280.249
      power (W): 238.041
      frequency (%): 75.8382
      frequency (Hz): 1.5926e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1700000000.000000
    -
      region: "stream-0.00-dgemm-6.00"
      hash: 0xaf4cafa3
      runtime (s): 17.8807
      sync-runtime (s): 17.8327
      package-energy (J): 4367.48
      dram-energy (J): 252.722
      power (W): 244.914
      frequency (%): 75.7953
      frequency (Hz): 1.5917e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1700000000.000000
    -
      region: "with some spaces"
      hash: 0xaf000fa3
      runtime (s): 17.8807
      sync-runtime (s): 17.8327
      package-energy (J): 4367.48
      dram-energy (J): 252.722
      power (W): 244.914
      frequency (%): 75.7953
      frequency (Hz): 1.5917e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1700000000.000000
    Unmarked Totals:
      runtime (s): 6.07865
      sync-runtime (s): 6.13216
      package-energy (J): 1043.75
      dram-energy (J): 72.5349
      power (W): 170.209
      frequency (%): 97.2362
      frequency (Hz): 2.04196e+09
      time-hint-network (s): 12
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 0
      requested-online-frequency: 2100000000.000000
    Epoch Totals:
      runtime (s): 0
      sync-runtime (s): 0
      package-energy (J): 0
      dram-energy (J): 0
      power (W): 0
      frequency (%): 0
      frequency (Hz): 0
      time-hint-network (s): 14
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 0
      time-hint-ignore (s): 0
    Application Totals:
      runtime (s): 541.507
      sync-runtime (s): 541.507
      package-energy (J): 83998.9
      dram-energy (J): 8218.69
      power (W): 155.121
      frequency (%): 0
      frequency (Hz): 0
      time-hint-network (s): 15
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      time-hint-ignore (s): 0
      geopmctl memory HWM (B): 147939328
      geopmctl network BW (B/sec): 0.0443208

  mcfly12:
    Final online freq map:
      0x0db5f27a: 1000000000.000000
      0x3a6c47e3: 1000000000.000000
      0x3c627f60: 1400000000.000000
      0x536c798f: 1000000000.000000
      0x725e8066: 2200000000.000000
      0x76244144: 1100000000.000000
      0x9851de3f: 1000000000.000000
      0xaddfa74f: 1500000000.000000
      0xaf4cafa3: 1700000000.000000
      0xce08ae24: 1300000000.000000
      0xe1242325: 1700000000.000000
      0xe50a9187: 1600000000.000000
      0xf9d11bbd: 1500000000.000000

    Regions:
    -
      region: "sleep"
      hash: 0x536c798f
      runtime (s): 275.837
      sync-runtime (s): 276.56
      package-energy (J): 34545.2
      dram-energy (J): 3811.41
      power (W): 124.91
      frequency (%): 50.0117
      frequency (Hz): 1.05024e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 5500
      requested-online-frequency: 1000000000.000000
    -
      region: "stream-0.35-dgemm-1.80"
      hash: 0x76244144
      runtime (s): 29.1273
      sync-runtime (s): 29.0865
      package-energy (J): 4968.14
      dram-energy (J): 713.884
      power (W): 170.806
      frequency (%): 53.9922
      frequency (Hz): 1.13384e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1100000000.000000
    -
      region: "stream-0.40-dgemm-1.20"
      hash: 0x0db5f27a
      runtime (s): 28.4557
      sync-runtime (s): 28.3904
      package-energy (J): 4697.92
      dram-energy (J): 745.835
      power (W): 165.476
      frequency (%): 50.3009
      frequency (Hz): 1.05632e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1000000000.000000
    -
      region: "stream-0.45-dgemm-0.60"
      hash: 0x9851de3f
      runtime (s): 27.2367
      sync-runtime (s): 27.1905
      package-energy (J): 4478.41
      dram-energy (J): 767.496
      power (W): 164.705
      frequency (%): 50.3711
      frequency (Hz): 1.05779e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1000000000.000000
    -
      region: "stream-0.30-dgemm-2.40"
      hash: 0xce08ae24
      runtime (s): 25.8974
      sync-runtime (s): 25.8082
      package-energy (J): 4672.56
      dram-energy (J): 636.411
      power (W): 181.05
      frequency (%): 61.493
      frequency (Hz): 1.29135e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1300000000.000000
    -
      region: "stream-0.50-dgemm-0.00"
      hash: 0x3a6c47e3
      runtime (s): 25.2686
      sync-runtime (s): 25.2523
      package-energy (J): 4412.83
      dram-energy (J): 776.741
      power (W): 174.75
      frequency (%): 50.9256
      frequency (Hz): 1.06944e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1000000000.000000
    -
      region: "stream-0.25-dgemm-3.00"
      hash: 0x3c627f60
      runtime (s): 24.5979
      sync-runtime (s): 24.5117
      package-energy (J): 4732.35
      dram-energy (J): 583.029
      power (W): 193.065
      frequency (%): 65.2297
      frequency (Hz): 1.36982e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1400000000.000000
    -
      region: "stream-0.20-dgemm-3.60"
      hash: 0xf9d11bbd
      runtime (s): 23.2737
      sync-runtime (s): 23.2129
      package-energy (J): 4694.3
      dram-energy (J): 527.13
      power (W): 202.228
      frequency (%): 68.9091
      frequency (Hz): 1.44709e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1500000000.000000
    -
      region: "stream-0.15-dgemm-4.20"
      hash: 0xaddfa74f
      runtime (s): 22.3063
      sync-runtime (s): 22.1766
      package-energy (J): 4650.85
      dram-energy (J): 466.899
      power (W): 209.719
      frequency (%): 68.7558
      frequency (Hz): 1.44387e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1500000000.000000
    -
      region: "stream-0.10-dgemm-4.80"
      hash: 0xe50a9187
      runtime (s): 20.3869
      sync-runtime (s): 20.3101
      package-energy (J): 4646.27
      dram-energy (J): 410.599
      power (W): 228.767
      frequency (%): 72.2666
      frequency (Hz): 1.5176e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1600000000.000000
    -
      region: "stream-0.05-dgemm-5.40"
      hash: 0xe1242325
      runtime (s): 19.2864
      sync-runtime (s): 19.1575
      package-energy (J): 4618.09
      dram-energy (J): 352.889
      power (W): 241.059
      frequency (%): 76.1363
      frequency (Hz): 1.59886e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1700000000.000000
    -
      region: "stream-0.00-dgemm-6.00"
      hash: 0xaf4cafa3
      runtime (s): 18.6718
      sync-runtime (s): 18.6886
      package-energy (J): 4600.85
      dram-energy (J): 321.137
      power (W): 246.185
      frequency (%): 75.8795
      frequency (Hz): 1.59347e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1700000000.000000
    -
      region: "with some spaces"
      hash: 0xaf000fa3
      runtime (s): 17.8807
      sync-runtime (s): 17.8327
      package-energy (J): 4367.48
      dram-energy (J): 252.722
      power (W): 244.914
      frequency (%): 75.7953
      frequency (Hz): 1.5917e+09
      time-hint-network (s): 0
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 500
      requested-online-frequency: 1700000000.000000
    Unmarked Totals:
      runtime (s): 1.15034
      sync-runtime (s): 1.11497
      package-energy (J): 226.452
      dram-energy (J): 28.9149
      power (W): 203.102
      frequency (%): 95.0914
      frequency (Hz): 1.99692e+09
      time-hint-network (s): 12
      time-hint-ignore (s): 0
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 0
      requested-online-frequency: 2200000000.000000
    Epoch Totals:
      runtime (s): 0
      sync-runtime (s): 0
      package-energy (J): 0
      dram-energy (J): 0
      power (W): 0
      frequency (%): 0
      frequency (Hz): 0
      time-hint-network (s): 14
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      count: 0
      time-hint-ignore (s): 0
    Application Totals:
      runtime (s): 541.505
      sync-runtime (s): 541.505
      package-energy (J): 85946.8
      dram-energy (J): 10142.6
      power (W): 158.718
      frequency (%): 0
      frequency (Hz): 0
      time-hint-network (s): 15
      time-hint-compute (s): 0
      time-hint-memory (s): 0
      time-hint-io (s): 0
      time-hint-serial (s): 0
      time-hint-parallel (s): 0
      time-hint-unknown (s): 0
      time-hint-unset (s): 0
      time-hint-ignore (s): 0
      geopmctl memory HWM (B): 145715200
      geopmctl network BW (B/sec): 0
"""

# First lines from a test_trace_runtimes integration test run
test_trace_data = """# geopm_version: 1.0.0+dev209g77e1ebb8
# start_time: Thu Oct 03 08:19:34 2019
# profile_name: test_trace_runtimes
# node_name: mcfly1
# agent: power_governor
TIME|EPOCH_COUNT|REGION_HASH|REGION_HINT|REGION_PROGRESS|REGION_COUNT|REGION_RUNTIME|CPU_ENERGY|DRAM_ENERGY|CPU_POWER|DRAM_POWER|FREQUENCY|CPU_CYCLES_THREAD|CPU_CYCLES_REFERENCE|CPU_CORE_TEMPERATURE|POWER_BUDGET
0.204296343|-1|0x00000000725e8066|0x0000000100000000|0|0|0|242598.262512207|31537.68629558909|nan|nan|2763636363.636364|44306314967139|44889410342036|60.18181818181818|0
0.2223164129999999|-1|0x00000000644f9787|0x0000000100000000|0|0|0|242603.1904296875|31538.00337322582|273.4682762313778|17.59580494025078|2800000000|44310710666495|44892733038308|61.72727272727273|150
0.2235088959999999|-1|0x00000000644f9787|0x0000000100000000|0|0|0|242603.6185302734|31538.02900799145|276.6301227683065|17.73233606304858|2800000000|44311004818789|44892952655048|61.77272727272727|150
0.228532226|-1|0x00000000644f9787|0x0000000100000000|0|0|0|242604.9693603516|31538.11309917797|276.9201297420731|17.65826611195596|2100000000|44312147188306|44893868186252|61.43181818181818|150
0.233763619|-1|0x00000000644f9787|0x0000000100000000|0|0|0|242605.6206665039|31538.18380840649|213.1977333899407|17.13213273088159|776136363.6363636|44312679661043|44894805921596|59.79545454545455|150
0.238758759|-1|0x00000000644f9787|0x0000000100000000|0|0|0|242606.1339111328|31538.24949749341|160.7312347320962|16.51782487831938|800000000|44313012842482|44895724971644|59.13636363636363|150
0.2436604579999999|-1|0x00000000644f9787|0x0000000100000000|0|0|0|242606.7144165039|31538.31727703442|114.1336448920131|16.01735740105445|1450000000|44313477045974|44896618887716|58.97727272727273|150
0.248572409|-1|0x00000000644f9787|0x0000000100000000|0|0|0|242607.4743041992|31538.39257915845|124.4617940659552|15.71872562446383|1877272727.272727|44314215214920|44897517408080|59.15909090909091|150
0.253558066|-1|0x00000000644f9787|0x0000000100000000|0|0|0|242608.3024291992|31538.46925457349|147.3800125521751|14.56509260230485|1850000000|44315033338888|44898435409136|59.29545454545455|150
0.2585734129999999|-1|0x00000000644f9787|0x0000000100000000|0|0|0|242609.1127319336|31538.54991253248|161.3588539038742|14.57640738621386|1650000000|44315800167450|44899353800036|59.13636363636363|150
0.263574746|-1|0x00000000644f9787|0x0000000100000000|0|0|0|242609.8388671875|31538.61998089185|158.0069444604364|14.60065683114463|1550000000|44316493094098|44900272907876|59.02272727272727|150
0.268616921|-1|0x00000000644f9787|0x0000000100000000|0|0|0|242610.5656738281|31538.70031841627|149.7814626529688|14.94329910436242|1600000000|44317174126049|44901195205784|58.93181818181818|150
"""


class TestIO(unittest.TestCase):
    def setUp(self):
        if 'assertCountEqual' not in dir(self):
            # Python 3 replaces assertItemsEqual with assertCountEqual
            self.assertCountEqual = self.assertItemsEqual
        self._test_directory = tempfile.mkdtemp()
        self._report_path = os.path.join(self._test_directory, 'geopmpy-io-test-raw-report')
        with open(self._report_path, 'w') as fid:
            fid.write(test_report_data)
        self._trace_path = os.path.join(self._test_directory, 'geopmpy-io-test-trace')
        with open(self._trace_path, 'w') as fid:
            fid.write(test_trace_data)

    def tearDown(self):
        shutil.rmtree(self._test_directory)

    def test_requested_online_frequency(self):
        report = geopmpy.io.RawReport(self._report_path)
        host_names = report.host_names()
        for nn in report.region_names(host_names[0]):
            if 'dgemm-0.00' in nn:
                stream_region_name = nn
            if 'stream-0.00' in nn:
                dgemm_region_name = nn
        stream_region = report.raw_region(host_names[0], stream_region_name)
        dgemm_region = report.raw_region(host_names[0], dgemm_region_name)
        self.assertLess(stream_region['requested-online-frequency'], dgemm_region['requested-online-frequency'])

        hosts = report.host_names()
        region_names = report.region_names(hosts[0])
        region = report.raw_region(hosts[1], region_names[1])
        region_runtime = region['runtime (s)']
        epoch = report.raw_epoch(hosts[0])
        runtime = epoch['runtime (s)']
        self.assertEqual(0, runtime)
        totals = report.raw_totals(hosts[1])
        total_runtime = totals['runtime (s)']
        self.assertLess(region_runtime, total_runtime)
        field_runtime = report.get_field(epoch, 'runtime')
        self.assertEqual(runtime, field_runtime)
        region = report.raw_region('mcfly11', 'sleep')
        self.assertTrue(type(region['frequency (Hz)']) is float)
        self.assertEqual(1e9, region['frequency (Hz)'])

    def test_report_cache(self):
        """ Test that a report is not read when it is cached.
        """
        spy_open = mock.Mock(wraps=open)

        def count_open(path):
            """ Count the number of times spy_open() has been called with path
            """
            return Counter([c[0][0] for c in spy_open.call_args_list])[path]

        with mock.patch('geopmpy.io.open', spy_open), self_cleaning_raw_report_collection(self._report_path):
            initial_call_count = count_open(self._report_path)

            geopmpy.io.RawReportCollection(self._report_path)
            self.assertEqual(initial_call_count, count_open(self._report_path))

            touch_file(self._report_path)
            geopmpy.io.RawReportCollection(self._report_path)
            self.assertEqual(initial_call_count * 2, count_open(self._report_path))

    def test_trace(self):
        """ Test that a trace file can be extracted to a dataframe.
        """
        app_output = geopmpy.io.AppOutput(traces=self._trace_path, do_cache=False)
        trace_df = app_output.get_trace_data(node_name='mcfly1')
        start_time = trace_df.index.get_level_values('start_time').unique()
        self.assertEqual(1, len(start_time))
        self.assertEqual('Thu Oct 03 08:19:34 2019', start_time[0])
        self.assertAlmostEqual(0.204296343, trace_df.iloc[0]['TIME'])
        self.assertAlmostEqual(242598.262512207, trace_df.iloc[0]['CPU_ENERGY'])
        self.assertAlmostEqual(0.268616921, trace_df.iloc[-1]['TIME'])
        self.assertAlmostEqual(242610.5656738281, trace_df.iloc[-1]['CPU_ENERGY'])

    def test_figure_of_merit(self):
        fom_report_path = os.path.join(os.path.dirname(__file__), 'test_io_experiment.report')

        # test that RawReport can find FOM
        rr = geopmpy.io.RawReport(path=fom_report_path)
        self.assertAlmostEqual(327780.0, rr.figure_of_merit())
        rrc = geopmpy.io.RawReportCollection(fom_report_path, do_cache=False)
        app_data = rrc.get_app_df()
        self.assertAlmostEqual(327780.0, app_data['FOM'].mean())

        # test that RawReport works when FOM is missing
        report = geopmpy.io.RawReport(self._report_path)
        self.assertTrue(report.figure_of_merit() is None)
        rrc = geopmpy.io.RawReportCollection(self._report_path, do_cache=False)
        self.assertFalse('FOM' in rrc.get_app_df().columns)

    def test_h5_name(self):
        # check that a set of reports produces the same hdf5 cache name
        # from any output dir relative path

        cwd = os.getcwd()
        basenames = ['file_a', 'file_b', 'file_c']
        for name in basenames:
            with open(name, 'w'):
                # touch file
                pass
        rel_set = [os.path.join('.', ff) for ff in basenames]
        abs_set = [os.path.join(cwd, ff) for ff in basenames]

        outdir = cwd
        rel_h5 = geopmpy.io.RawReportCollection.make_h5_name(rel_set, outdir)
        abs_h5 = geopmpy.io.RawReportCollection.make_h5_name(abs_set, outdir)
        self.assertEqual(rel_h5, abs_h5)

        for ff in basenames:
            os.unlink(ff)

    def test_raw_report_collection_region(self):
        rrc = geopmpy.io.RawReportCollection(self._report_path, do_cache=False)
        df = rrc.get_df()
        sleep_region_mcfly11 = {'GEOPM Version': '1.0.0+dev30g4cccfda',
                                'Start Time': 'Thu May 30 14:38:17 2019',
                                'Profile': 'test_fm_stream_dgemm_mix',
                                'Agent': 'frequency_map',
                                'host': 'mcfly11',
                                'region': 'sleep',
                                'hash': 0x536c798f,
                                'runtime (s)': 275.998,
                                'sync-runtime (s)': 276.581,
                                'package-energy (J)': 33573.9,
                                'dram-energy (J)': 2835.43,
                                'power (W)': 121.389,
                                'frequency (%)': 50.0846,
                                'frequency (Hz)': 1e+09,
                                'time-hint-network (s)': 0,
                                'time-hint-ignore (s)': 0,
                                'time-hint-compute (s)': 0,
                                'time-hint-memory (s)': 0,
                                'time-hint-io (s)': 0,
                                'time-hint-serial (s)': 0,
                                'time-hint-parallel (s)': 0,
                                'time-hint-unknown (s)': 0,
                                'time-hint-unset (s)': 0,
                                'count': 5500,
                                'requested-online-frequency': 1000000000.000000}
        self.assertEqual(set(sleep_region_mcfly11.keys()), set(df.columns))
        actual_sleep_region = df.loc[(df['host'] == 'mcfly11') &
                                     (df['region'] == 'sleep')].to_dict('records')[0]
        self.assertEqual(sleep_region_mcfly11, actual_sleep_region)

    def test_raw_report_collection_app_totals(self):
        rrc = geopmpy.io.RawReportCollection(self._report_path, do_cache=False)
        df = rrc.get_app_df()
        app_totals_mcfly12 = {'GEOPM Version': '1.0.0+dev30g4cccfda',
                              'Start Time': 'Thu May 30 14:38:17 2019',
                              'Profile': 'test_fm_stream_dgemm_mix',
                              'Agent': 'frequency_map',
                              'host': 'mcfly12',
                              'runtime (s)': 541.505,
                              'sync-runtime (s)': 541.505,
                              'package-energy (J)': 85946.8,
                              'dram-energy (J)': 10142.6,
                              'power (W)': 158.718,
                              'frequency (%)': 0,
                              'frequency (Hz)': 0,
                              'time-hint-network (s)': 15,
                              'time-hint-compute (s)': 0,
                              'time-hint-memory (s)': 0,
                              'time-hint-io (s)': 0,
                              'time-hint-serial (s)': 0,
                              'time-hint-parallel (s)': 0,
                              'time-hint-unknown (s)': 0,
                              'time-hint-unset (s)': 0,
                              'time-hint-ignore (s)': 0,
                              'geopmctl memory HWM (B)': 145715200,
                              'geopmctl network BW (B/sec)': 0}
        self.assertEqual(set(app_totals_mcfly12.keys()), set(df.columns))
        actual_totals= df.loc[(df['host'] == 'mcfly12')].to_dict('records')[0]
        self.assertEqual(app_totals_mcfly12, actual_totals)

    def test_raw_report_collection_epoch_totals(self):
        rrc = geopmpy.io.RawReportCollection(self._report_path, do_cache=False)
        df = rrc.get_epoch_df()
        epoch_totals_mcfly12 = {'GEOPM Version': '1.0.0+dev30g4cccfda',
                                'Start Time': 'Thu May 30 14:38:17 2019',
                                'Profile': 'test_fm_stream_dgemm_mix',
                                'Agent': 'frequency_map',
                                'host': 'mcfly12',
                                'runtime (s)': 0,
                                'sync-runtime (s)': 0,
                                'package-energy (J)': 0,
                                'dram-energy (J)': 0,
                                'power (W)': 0,
                                'frequency (%)': 0,
                                'frequency (Hz)': 0,
                                'time-hint-network (s)': 14,
                                'time-hint-compute (s)': 0,
                                'time-hint-memory (s)': 0,
                                'time-hint-io (s)': 0,
                                'time-hint-serial (s)': 0,
                                'time-hint-parallel (s)': 0,
                                'time-hint-unknown (s)': 0,
                                'time-hint-unset (s)': 0,
                                'count': 0,
                                'time-hint-ignore (s)': 0}
        self.assertEqual(set(epoch_totals_mcfly12.keys()), set(df.columns))
        actual_totals= df.loc[(df['host'] == 'mcfly12')].to_dict('records')[0]
        self.assertEqual(epoch_totals_mcfly12, actual_totals)

    def test_raw_report_collection_unmarked_totals(self):
        rrc = geopmpy.io.RawReportCollection(self._report_path, do_cache=False)
        df = rrc.get_unmarked_df()
        unmarked_mcfly11 = {'GEOPM Version': '1.0.0+dev30g4cccfda',
                            'Start Time': 'Thu May 30 14:38:17 2019',
                            'Profile': 'test_fm_stream_dgemm_mix',
                            'Agent': 'frequency_map',
                            'host': 'mcfly11',
                            'runtime (s)': 6.07865,
                            'sync-runtime (s)': 6.13216,
                            'package-energy (J)': 1043.75,
                            'dram-energy (J)': 72.5349,
                            'power (W)': 170.209,
                            'frequency (%)': 97.2362,
                            'frequency (Hz)': 2.04196e+09,
                            'time-hint-network (s)': 12,
                            'time-hint-ignore (s)': 0,
                            'time-hint-compute (s)': 0,
                            'time-hint-memory (s)': 0,
                            'time-hint-io (s)': 0,
                            'time-hint-serial (s)': 0,
                            'time-hint-parallel (s)': 0,
                            'time-hint-unknown (s)': 0,
                            'time-hint-unset (s)': 0,
                            'count': 0,
                            'requested-online-frequency': 2100000000.000000}
        self.assertEqual(set(unmarked_mcfly11.keys()), set(df.columns))
        actual = df.loc[(df['host'] == 'mcfly11')].to_dict('records')[0]
        self.assertEqual(unmarked_mcfly11, actual)


if __name__ == '__main__':
    unittest.main()
